/*
 * burtBackup.c
 *
 *    This file implements the backup command.  This is the nuts and bolts
 *    of Burt's backup services.
 *
 * Copyright (C) 1995 - 1999 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifndef lint
static char rcsid[] = "$Id: burtBackup.c,v 1.45 1999/05/06 22:19:16 jmelski Exp $";
#endif
#define DEBUG 0

#include <burtInt.h>
#include <signal.h>
#include <sys/times.h>

typedef struct Burt_BackupMetaControlBlock {
  Tcl_Interp *interp;            /* Controlling interpreter     */
  Tcl_Channel outputChannel;     /* Channel to write output.    */
  Tcl_Channel logChannel;        /* Channel to write log data.  */
} Burt_BackupMetaControlBlock;

typedef struct Burt_BackupControlBlock {
  Burt_Packet packet;            /* Packet for this backup. */
  Tcl_TimerToken sentryTimer;    /* Token to the sentry timer. */
  unsigned int sentry;           /* Holds number of bytes read from
				  * inputChannel at last sentryTimer */
  Tcl_Channel inputChannel;      /* Channel to read data from. */
  Tcl_Channel stderrChannel;     /* Channel to read stderr.    */
  char *backupMonitor;           /* Name of user def'd proc to send stderr 
                                    * to check for errors, or just dialog. */
} Burt_BackupControlBlock;

/*
 * This structure describes per-instance state of a pipe based channel.
 * Lifted, as is, from the Tcl source.
 */
typedef struct PipeState {
  Tcl_Channel channel;  /* Channel associated with this file. */
  TclFile inFile;       /* Output from pipe. */
  TclFile outFile;	/* Input to pipe. */
  TclFile errorFile;	/* Error output from pipe. */
  int numPids;	        /* How many processes are attached to this pipe? */
  int *pidPtr;	        /* The process IDs themselves. Allocated by
		         * the creator of the pipe. */
  int isNonBlocking;	/* Nonzero when the pipe is in nonblocking mode.
                         * Used to decide whether to wait for the children
                         * at close time. */
} PipeState;

/* Functions private to the backup module */
static int  AbortBackup(void);
static int  AppendBackup(Tcl_Interp *interp, 
			 char *outputChannelName, 
			 char *logChannelName, 
			 int numChannels);
static int  StartBackup(Tcl_Interp *interp, 
			char *outputChannelName, 
			char *logChannelName, 
			int numChannels, 
			char *label);
static int  StopBackup(void);
static int  StartBackupChannel(void);
static int  EvaluateUserBackupProc(Tcl_Interp *interp, 
				   Burt_BackupControlBlock *backupCB);
static void CloseBackupChannel(Burt_BackupControlBlock *backupCB);
static void ProcessBackupInputChannel(ClientData clientData, int mask);
static void ProcessBackupStderrChannel(ClientData clientData, int mask);
static void ProcessBackupSentry(ClientData clientData);
static void FlushBackupStderrChannel(Burt_BackupControlBlock *backupCB);
static int  BurtBackupIsComplete(void);
static int  LoadNextBackupTape(Tcl_Channel mtChannel, Tcl_Channel logChannel);
static void FormatBackupProcError(Tcl_Interp *interp, 
				  char *procName, 
				  char *host, 
				  char *atom, 
				  char *level, 
				  char *errorMessage);

static void BurtEnableBackupEvents(Burt_BackupControlBlock *backupCB);
static void BurtDisableBackupEvents(Burt_BackupControlBlock *backupCB);
static int  GetUserBackupMonitorProc(Tcl_Interp *interp, 
				     Burt_BackupControlBlock *backupCB);

/* Data private to the backup module */
static Tcl_HashTable backupCBTable;      /* Holds the currently running CB's */
static Burt_BackupMetaControlBlock backupMCB;     /* Data common to all CB's */
static char *savedBackupStatus;     /* Saves backup status during readchecks */


/*
 *----------------------------------------------------------------------
 *
 * Burt_BackupObjCmd --
 *
 *	This procedure implements the Tcl command backup.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *      Depending on input, may start, stop, abort or append a backup
 *----------------------------------------------------------------------
 */
int Burt_BackupObjCmd(ClientData cdata, Tcl_Interp *interp, 
		      int objc, Tcl_Obj *CONST objv[])
{
  char *outputChannelName, *logChannelName, *label;
  int numChannels;
  static char *options[] = { "abort", "append", "start", "stop", NULL };
  enum options { BACKUP_ABORT, BACKUP_APPEND, BACKUP_START, BACKUP_STOP };
  int index;

  if (objc < 2 || objc > 6)
  {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
    return(TCL_ERROR);
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], options, 
			  "option", 0, &index) != TCL_OK)
    return(TCL_ERROR);

  switch ((enum options) index)
  {
    case BACKUP_ABORT:
    {
      if (AbortBackup() == TCL_ERROR)
	return(TCL_ERROR);
      break;
    }
    case BACKUP_APPEND:
    {
      if (!BurtBackupIsComplete())
      {
	Tcl_SetResult(interp, "BURT: Backup already running", TCL_STATIC);
	return(TCL_ERROR);
      }

      if (objc != 6)
      {
	Tcl_WrongNumArgs(interp, 2, objv, 
			 "channels outputchannel logchannel label");
	return(TCL_ERROR);
      }
      
      if (Tcl_GetIntFromObj(interp, objv[2], &numChannels) == TCL_ERROR)
	return(TCL_ERROR);

      outputChannelName = Tcl_GetStringFromObj(objv[3], NULL);
      logChannelName    = Tcl_GetStringFromObj(objv[4], NULL);

      if (AppendBackup(interp, 
		       outputChannelName, 
		       logChannelName, 
		       numChannels) == TCL_ERROR)
	return(TCL_ERROR);

      break;
    } 
    case BACKUP_START:
    {
      if (!BurtBackupIsComplete())
      {
	Tcl_SetResult(interp, "BURT: Backup already running", TCL_STATIC);
	return(TCL_ERROR);
      }

      if (objc != 6)
      {
	Tcl_WrongNumArgs(interp, 2, objv, 
			 "channels outputchannel logchannel label");
	return(TCL_ERROR);
      }

      if (Tcl_GetIntFromObj(interp, objv[2], &numChannels) == TCL_ERROR)
	return(TCL_ERROR);

      outputChannelName = Tcl_GetStringFromObj(objv[3], NULL);
      logChannelName    = Tcl_GetStringFromObj(objv[4], NULL);
      label             = Tcl_GetStringFromObj(objv[5], NULL);

      if (StartBackup(interp, 
		      outputChannelName, 
		      logChannelName, 
		      numChannels, 
		      label) == TCL_ERROR)
	return(TCL_ERROR);

      break;
    }
    case BACKUP_STOP:
    {
      if (StopBackup() == TCL_ERROR)
	return(TCL_ERROR);
      break;
    }
  }
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * AbortBackup --
 *
 *	This procedure aborts all running backup channels by repeatedly
 *      calling CloseBackupChannel, stamps the log, sets the run state to 
 *      ABORT, and sets burt_complete
 *
 * Side effects:
 *      All backup channels are closed; burt_complete is set to 2, 
 *      backupCBTable is deleted, run state is set to ABORT.
 *----------------------------------------------------------------------
 */
static int
AbortBackup(void)
{
  Burt_BackupControlBlock *backupCB;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;

  if (BurtBackupIsComplete())
  {
    Tcl_SetResult(backupMCB.interp, "BURT: No backup running ", TCL_STATIC);
    Tcl_AppendResult(backupMCB.interp, BurtTime(), "\n", NULL);
    return(TCL_ERROR);
  }
  
  Tcl_Write(backupMCB.logChannel, "BURT: Received abort command ", -1);
  Tcl_Write(backupMCB.logChannel, BurtTime(), -1);
  Tcl_Write(backupMCB.logChannel, "\n", -1);

  BurtStatusSetRunState(BACKUP | ABORT);

  entryPtr = Tcl_FirstHashEntry(&backupCBTable, &searchPtr);
  while (entryPtr != NULL)
  {
    backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
    CloseBackupChannel(backupCB);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }

  Tcl_ResetResult(backupMCB.interp);
  Tcl_AppendResult(backupMCB.interp, 
		   "BURT: Backups aborted ", BurtTime(), "\n", NULL);
  Tcl_Write(backupMCB.logChannel, Tcl_GetStringResult(backupMCB.interp), -1);
  BurtStatusSetCompleteTime(time((time_t *) NULL));
  Tcl_DeleteHashTable(&backupCBTable);
  Tcl_SetVar(backupMCB.interp, "burt_complete", "2", TCL_GLOBAL_ONLY);
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * StartBackup --
 *
 *	This procedure starts a backup with the specified output and log 
 *      channels, number of channels and tape label.  Called when the user
 *      types "backup start ..."
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *      Starts a backup.
 *	
 *----------------------------------------------------------------------
 */
int
StartBackup (Tcl_Interp *interp, char *outputChannelName, 
	     char *logChannelName, int numChannels, char *label)
{
  Tcl_Channel outputChannel, logChannel;
  Tcl_Obj *resultPtr;
  int mode;
  int nbytes;
  char msg[32];
  int startedChannels;
  Burt_TapeLabel tapeLabel;

  /* Make sure that the tape name is not too long */
  if (strlen(label) >= BURT_MAX_LABELLEN)
  {
    resultPtr = Tcl_GetObjResult(interp);
    sprintf(msg, "%d", strlen(label));
    Tcl_SetStringObj(resultPtr, "BURT: Tape label is too long (", -1);
    Tcl_AppendStringsToObj(resultPtr, msg, 
			   " chars, should be less than ", NULL);
    sprintf(msg, "%d", BURT_MAX_LABELLEN);
    Tcl_AppendStringsToObj(resultPtr, msg, " chars)", BurtTime(), NULL);
    return(TCL_ERROR);
  }

  /* Make sure something has been scheduled for backup */
  if (BurtScheduleLength() == 0) 
  {
    Tcl_SetResult(interp, "BURT: Nothing scheduled for backup", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* Get the output channel handle and its mode */
  outputChannel = Tcl_GetChannel(interp, outputChannelName, &mode);
  if (outputChannel == NULL)
    return(TCL_ERROR);

  /* 
   * The output channel must be both readable and writable, so we can 
   * write data to it (duh) and so we can read data from it for readchecks
   */
  if (((mode & TCL_WRITABLE) == 0) || ((mode & TCL_READABLE) == 0))
  {
    Tcl_SetResult(interp, "BURT: Output channel is not both "
		  "readable and writable", TCL_STATIC);
    return(TCL_ERROR);
  }     

  /* Get the log channel handle */
  logChannel = Tcl_GetChannel(interp, logChannelName, &mode);
  if (logChannel == NULL)
    return(TCL_ERROR);

  /* The log channel should be writable */
  if ((mode & TCL_WRITABLE) == 0) 
  {
    Tcl_SetResult(interp, "BURT: Log channel is not writable", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* Everything is OK; store the handles and interp pointer in the module */
  backupMCB.interp = interp;
  backupMCB.outputChannel = outputChannel;
  backupMCB.logChannel = logChannel;

  /* 
   * Rewind the tape in anticipation of writing it.
   * NOTE:  If Burt is to be truly data-destination independant, then a failed
   * rewind maybe should NOT cause an error ??  It may depend on how much of 
   * the "Mt-ness" is pulled down to the Mt layer.  For example, should this
   * function, at least the parts about postitioning the tape be in burtBackup,
   * or in burtMtIO?  Perhaps this fxn should call a "PrepareTargetForBackup"
   * fxn, which wouuld be differently implemented in MtIO than in DiskIO ...
   */
  if (BurtMtRewind(outputChannel) != MT_OK)
  {
    Tcl_Write(logChannel, "BURT: WARNING, unable to rewind tape ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
  }

  /* Initialize and write the new tape label */
  SetupTapeLabel(&tapeLabel, label, 1);
  nbytes = BurtMtWriteLabel(outputChannel, &tapeLabel);
  if (nbytes == MT_ERROR)
  {
    Tcl_SetResult(interp, "BURT: Could not write tape label", TCL_STATIC);
    return(TCL_ERROR); 
  }

  /* Initialize the readtape module, for interrupted readchecks */
  BurtReadtapeInit();

  /* Reset the status stuff, then set it to appropriate info */
  BurtStatusReset();
  BurtStatusSetRunState(BACKUP | RUN);
  BurtStatusSetTapeLabel(label);
  BurtStatusSetScheduled(BurtScheduleLength());
  BurtStatusSetTapes(1);
  BurtStatusSetStartTime(time((time_t *) NULL));
  BurtStatusSetBytes(BurtStatusGetBytes() + nbytes);

  /* 
   * Initialize the backupCBTable, which holds pointers to each running item
   */
  Tcl_InitHashTable(&backupCBTable, TCL_STRING_KEYS);

  /* 
   * Start the requested number of backup channels.
   * If there are problems, we may start less than the requested number
   */
  startedChannels = 0;
  while (startedChannels < numChannels && BurtScheduleLength() > 0)
  {
    /*
     * If an error occurs when trying to start a channel, try again with
     * a different schedule item; continue in this fashion until either the
     * correct number of channels has been started or we run out of items to
     * try to start
     */
    if (StartBackupChannel() == TCL_ERROR)
    {
      /* StartBackupChannel fills in an error, we just have to log it */
      Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
      Tcl_Flush(logChannel);
      BurtStatusSetFinished(BurtStatusGetFinished() + 1);
    }
    else
    {
      startedChannels++;
    }
  }
  /* Update the status information to reflect the number of channels started */
  BurtStatusSetChannels(startedChannels);

  /* Set burt_complete to indicate that backups are going; the Tcl script 
   * should watch for changes to burt_complete (with vwait), which signify that
   * backups have completed
   */
  Tcl_SetVar(interp, "burt_complete", "0", TCL_GLOBAL_ONLY);

  resultPtr = Tcl_GetObjResult(interp);
  sprintf(msg, "%d", startedChannels);
  Tcl_SetStringObj(resultPtr, "BURT: Started ", -1);
  Tcl_AppendStringsToObj(resultPtr, " ", msg, " channels ", BurtTime(), NULL);
  
  return(TCL_OK);
}
	     

/*
 *----------------------------------------------------------------------
 *
 * StopBackup --
 *
 *	This procedure stops a backup when a user types "backup stop"
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *      Stops a backup by setting the runstate to BACKUP | STOP, which
 *      allows already started items to complete, but will not start 
 *      new items.
 *
 *----------------------------------------------------------------------
 */
static int
StopBackup (void)
{
  if (BurtBackupIsComplete())
  {
    Tcl_SetResult(backupMCB.interp, "BURT: No backup running", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  Tcl_Write(backupMCB.logChannel, "BURT: Received stop command ", -1);
  Tcl_Write(backupMCB.logChannel, BurtTime(), -1);
  Tcl_Write(backupMCB.logChannel, "\n", -1);
      
  /* Set runstate to STOP; running channels will finish, no more will start */
  BurtStatusSetRunState(BACKUP | STOP);
  
  Tcl_ResetResult(backupMCB.interp);
  Tcl_AppendResult(backupMCB.interp, 
		   "BURT: Backup stopped, running channels will finish ", 
		   BurtTime(),
		   NULL);
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * AppendBackup --
 *
 *	This procedure starts a backup in "append" mode, which means that
 *      the data will be written after any existing data on the tape.  It 
 *      goes through a lengthy sequence of shenanigans to position the tape
 *      and ensure that it can really be appended to, and if any of those
 *      fail, then it won't append.
 *
 * Results:
 *	Returns TCL_OK if it successfully starts the backup in append mode,
 *      else it returns TCL_ERROR.  Fills interp->result with info on error.
 *
 * Side effects:
 *      Starts a backup in append mode
 *      
 *----------------------------------------------------------------------
 */
int 
AppendBackup (Tcl_Interp *interp, char *outputChannelName, 
	      char *logChannelName, int numChannels)
{
  Tcl_Channel outputChannel, logChannel;
  Tcl_Obj *resultPtr;
  int mode;
  int nbytes;
  char msg[32];
  int startedChannels;
  Burt_TapeLabel tapeLabel;
  Burt_Packet packet;

  /* 
   * Make sure there's something to backup by checking the schedule length
   */
  if (BurtScheduleLength() == 0) 
  {
    Tcl_SetResult(interp, "BURT: Nothing scheduled for backup", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* Get the output channel handle */
  outputChannel = Tcl_GetChannel(interp, outputChannelName, &mode);
  if (outputChannel == NULL)
    return(TCL_ERROR);

  /* 
   * The output channel must be both writable and readable, so that we can
   * write data to it, and so we can read it back for interrupted readchecks
   */
  if (((mode & TCL_WRITABLE) == 0) || ((mode & TCL_READABLE) == 0))
  {
    Tcl_SetResult(interp, "BURT: Output channel is not both "
		  "readable and writable", TCL_STATIC);
    return(TCL_ERROR);
  }     

  /* Get the log channel handle */
  logChannel = Tcl_GetChannel(interp, logChannelName, &mode);
  if (logChannel == NULL)
    return(TCL_ERROR);

  /* Make sure the log channel is writable */
  if ((mode & TCL_WRITABLE) == 0) 
  {
    Tcl_SetResult(interp, "BURT: Log channel is not writable", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* 
   * Good, the output and log channels are set up correctly.  Store the handles
   * and a pointer to the interpreter in the module data.
   */
  backupMCB.interp = interp;
  backupMCB.outputChannel = outputChannel;
  backupMCB.logChannel = logChannel;

  /* Rewind the tape in anticipation of writing it */
  if (BurtMtRewind(outputChannel) != MT_OK)
  {
    Tcl_Write(logChannel, "BURT: WARNING, unable to rewind tape ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
  }

  Tcl_Write(logChannel, "BURT: Verifying BURT version on tape... ", -1);
  Tcl_Write(logChannel, BurtTime(), -1);
  Tcl_Write(logChannel, "\n", -1);
  Tcl_Flush(logChannel);

  /* Read the tape label */
  if (BurtMtReadLabel(outputChannel, &tapeLabel) != MT_OK)
  {
    Tcl_SetResult(interp, "BURT: Unable to read label", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* Make sure tape was written by BURT v1.5 or greater; v1.5 was the 
   * first version to support backup appending
   */
  if (tapeLabel.magic < 1500)
  {
    Tcl_SetResult(interp, "BURT: Tape written by a BURT version which "
		  "did not support append", TCL_STATIC);
    return(TCL_ERROR);
  }

  if (BurtMtSafeRewind(outputChannel) != MT_OK)
  {
    Tcl_SetResult(interp, "BURT: Unable to rewind tape", TCL_STATIC);
    return(TCL_ERROR);
  }

  Tcl_Write(logChannel, "BURT: Seeking end of media... ", -1);
  Tcl_Write(logChannel, BurtTime(), -1);
  Tcl_Write(logChannel, "\n", -1);
  Tcl_Flush(logChannel);

  /* Go to the end of the tape */
  if (BurtMtEOM(outputChannel) != MT_OK)
  {
    Tcl_SetResult(interp, "BURT: Unable to find EOM", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  Tcl_Write(logChannel, "BURT: Located end of media ", -1);
  Tcl_Write(logChannel, BurtTime(), -1);
  Tcl_Write(logChannel, "\n", -1);
  Tcl_Write(logChannel, "BURT: Seeking END_OF_BACKUPSET packet... ", -1);
  Tcl_Write(logChannel, BurtTime(), -1);
  Tcl_Write(logChannel, "\n", -1);
  Tcl_Flush(logChannel);

  /* Backspace two filemarks. This positions the tape head before the EOF
   * that preceeds the END_OF_BACKUPSET packet (hopefully!); if we find an
   * END_OF_BACKUPSET packet there, then the tape can is appendable
   */
  if (BurtMtBSF(outputChannel, 2) != MT_OK)
  {
    Tcl_SetResult(interp, "BURT: Unable to backspace filemarks", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  /* Read the packet found at that position */
  nbytes = BurtMtRead(outputChannel, &packet);
  if (nbytes != sizeof(Burt_Packet))
  {
    Tcl_SetResult(interp, "BURT: Unable to read packet", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  /* Make sure it's the END_OF_BACKUPSET packet */
  if (packet.blockSize != END_OF_BACKUPSET)
  {
    Tcl_SetResult(interp, "BURT: Unable to find END_OF_BACKUPSET", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  /* Backspace one filemark, put the head before EOF before END_OF_BACKUPSET */
  if (BurtMtBSF(outputChannel, 1) != MT_OK)
  {
    Tcl_SetResult(interp, "BURT: Unable to backspace filemarks", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  /* If all of the above stuff goes through OK, then we are set to append */
  Tcl_Write(logChannel, "BURT: Found END_OF_BACKUPSET, ready to append  ", -1);
  Tcl_Write(logChannel, BurtTime(), -1);
  Tcl_Write(logChannel, "\n", -1);
  Tcl_Flush(logChannel);

  /* Initialize the readtape module, for interrupted readchecks */
  BurtReadtapeInit();

  /* Reset the status stuff, then set it to appropriate info */
  BurtStatusReset();
  BurtStatusSetRunState(BACKUP | RUN);
  BurtStatusSetScheduled(BurtScheduleLength());
  BurtStatusSetTapes(atoi(tapeLabel.tape));
  BurtStatusSetStartTime(time((time_t *) NULL));
  BurtStatusSetTapeLabel(tapeLabel.label);

  /* 
   * Initialize the backupCBTable, which holds pointers to each running item
   */
  Tcl_InitHashTable(&backupCBTable, TCL_STRING_KEYS);
  /* 
   * Start the specified number of backup channels, provided there are enough
   * schedule items to fill the request
   */
  startedChannels = 0;
  while (startedChannels < numChannels && BurtScheduleLength() > 0)
  {
    /*
     * If we get an error back when we try to start a channel, we should
     * try to start it again (with a different schedule item), until we have
     * either started the correct number of channels or run out of schedule 
     * items to try to start.
     */
    if (StartBackupChannel() == TCL_ERROR)
    {
      /* StartBackupChannel fills in an error, we just have to log it */
      Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
      Tcl_Flush(logChannel);
      BurtStatusSetFinished(BurtStatusGetFinished() + 1);
    }
    else
    {
      startedChannels++;
    }
  }
  /* Now we know how many channels we have started, so update the status */
  BurtStatusSetChannels(startedChannels);

  /* Set burt_complete to indicate that backups are going; the Tcl script 
   * should watch burt_complete (with vwait) for changes, which indicate 
   * that backups have finished
   */
  Tcl_SetVar(interp, "burt_complete", "0", TCL_GLOBAL_ONLY);

  resultPtr = Tcl_GetObjResult(interp);
  sprintf(msg, "%d", startedChannels);
  Tcl_SetStringObj(resultPtr, "BURT: Started ", -1);
  Tcl_AppendStringsToObj(resultPtr, " ", msg, " channels ", BurtTime(), NULL);
  
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * StartBackupChannel --
 *
 *	This procedure sets up a new channel handler for a scheduled item.
 *      After getting the next item from the schedule table it arranges 
 *      for the channels to be processed whenever the given channel has 
 *      data waiting to be read.  During this time a Burt_BackupControlBlock 
 *      is built, which is passed along to the channel handlers.
 *
 * Results:
 *	Returns TCL_OK on succes, or TCL_ERROR on failure.  Fills in
 *      interp->result with info upon error.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
static int 
StartBackupChannel(void)
{
  Burt_ScheduleItem *scheduleItemPtr;
  Burt_BackupControlBlock *backupCB;
  Tcl_HashEntry *entryPtr;
  Tcl_Obj *resultPtr;
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  int new;

  scheduleItemPtr = 
    BurtScheduleGetNext(backupMCB.interp, backupMCB.logChannel);

  if (scheduleItemPtr != NULL) 
  {
    backupCB = 
      (Burt_BackupControlBlock *)malloc(sizeof(Burt_BackupControlBlock));
    if (backupCB == NULL) 
    {
      Tcl_SetResult(backupMCB.interp, 
		    "BURT: Out of memory in StartBackupChannel", TCL_STATIC);
      return(TCL_ERROR);
    }
	
    /* Set the packet header information to reasonable values */
    backupCB->packet.magic     = BURT_MAGIC_NUMBER;
    backupCB->packet.blockNum  = 0;
    backupCB->packet.blockSize = 0;
    backupCB->packet.totalSize = 0;
    backupCB->packet.checksum  = 0;
    strncpy(backupCB->packet.host, scheduleItemPtr->host, BURT_MAX_HOSTLEN);
    strncpy(backupCB->packet.atom, scheduleItemPtr->atom, BURT_MAX_ATOMLEN);
    strncpy(backupCB->packet.type, scheduleItemPtr->type, BURT_MAX_TYPELEN);
    strncpy(backupCB->packet.level, scheduleItemPtr->level, BURT_MAX_LEVELLEN);
    free((void *) scheduleItemPtr);

    /* Fill in the backup monitor proc field of the backup control block */
    if (GetUserBackupMonitorProc(backupMCB.interp, backupCB) == TCL_ERROR)
      return(TCL_ERROR);

    /* Fill in the inputChannel and stderrChannel fields of the backupCB */
    if (EvaluateUserBackupProc(backupMCB.interp, backupCB) == TCL_ERROR) 
      return(TCL_ERROR);

    /* Add the backupCB to the backupCBTable */
    BurtSetupHashKey(key, 
		     backupCB->packet.host, 
		     backupCB->packet.atom,
		     backupCB->packet.level, 
		     backupCB->packet.type);
    entryPtr = Tcl_CreateHashEntry(&backupCBTable, key, &new);

    if (new)
    {
      Tcl_SetHashValue(entryPtr, backupCB);
    }
    else
    {
      resultPtr = Tcl_NewStringObj("BURT: ", -1);
      Tcl_AppendStringsToObj(resultPtr, "Item ",
			     backupCB->packet.host, " ",
			     backupCB->packet.atom, " ",
			     backupCB->packet.level, " ",
			     backupCB->packet.type, " ",
			     "already started"
			     " ", BurtTime(), NULL);
      Tcl_SetObjResult(backupMCB.interp, resultPtr);
      return (TCL_ERROR);
    }

    BurtEnableBackupEvents(backupCB);

    BurtStatusSetStarted(BurtStatusGetStarted() + 1);

    Tcl_Write(backupMCB.logChannel, "BURT: Started ", -1);
    Tcl_Write(backupMCB.logChannel, backupCB->packet.host, -1);
    Tcl_Write(backupMCB.logChannel, " ", -1);
    Tcl_Write(backupMCB.logChannel, backupCB->packet.atom, -1);
    Tcl_Write(backupMCB.logChannel, " ", -1);
    Tcl_Write(backupMCB.logChannel, backupCB->packet.level, -1);
    Tcl_Write(backupMCB.logChannel, " ", -1);
    Tcl_Write(backupMCB.logChannel, backupCB->packet.type, -1);
    Tcl_Write(backupMCB.logChannel, " ", -1);
    Tcl_Write(backupMCB.logChannel, BurtTime(), -1);
    Tcl_Write(backupMCB.logChannel, "\n", -1);
    Tcl_Flush(backupMCB.logChannel);
  }
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * CloseBackupChannel --
 *
 *	This procedure closes a backup channel, it deletes all of the
 *      event handlers associated with the backup, frees the backupCB,
 *      and attempts to start a new backup channel to replace the one
 *      which was just closed.  It honors the run state, so if burt is
 *      not in the "RUN" run state it will only do the close, and skip
 *      the starting of a  new channel.  If there are no more backups
 *      to start up, it will write out the final END_OF_BACKUPSET to the
 *      tape to make the end of the run.
 *
 * Results:
 *
 * Side effects:
 *     New backup channel may be started.
 *----------------------------------------------------------------------
 */

static void 
CloseBackupChannel(Burt_BackupControlBlock *backupCB)
{
  Tcl_ChannelType *chanTypePtr;	/* The type of that channel. */
  PipeState *pipePtr;     /* The state of the pipe handed back user. */
  register int i;
  int done;
  char *cleanupProc;
  char logBuffer[1024];
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  Tcl_HashEntry *entryPtr;

  BurtStatusSetFinished(BurtStatusGetFinished() + 1);

  /*
   * If the user gave us pipes, kill the associated processes
   * before we close up the pipes, this must be done to prevent
   * the processes (namely rsh) from hanging.
   *
   * IMPORTANT DEVELOPMENT NOTE:  If we didn't poke directly into the PipeState
   * structure here, we would not need to #include tclInt.h, which would make
   * config/compile easier on _everybody_.  We could avoid this direct poke by
   * using the Tcl function "pid".  However, that would require a Tcl_Eval 
   * followed by a Tcl_ParseList and a bunch of string to int conversions.  The
   * overhead may be significant (remember, this happens pretty much every time
   * a channel is closed!)
   */
  
  chanTypePtr = Tcl_GetChannelType(backupCB->inputChannel);
  if (strcmp(chanTypePtr->typeName, "pipe") == 0) 
  {
    pipePtr = (PipeState *) Tcl_GetChannelInstanceData(backupCB->inputChannel);
    for (i = 0; i < pipePtr->numPids; i++) 
    {
      kill((int)(pipePtr->pidPtr[i]), SIGKILL);
    }
  }

  /* 
   * Delete all the event handlers. Close up input and stderr channels.
   */

  Tcl_DeleteChannelHandler(backupCB->inputChannel, ProcessBackupInputChannel,
			   backupCB);
  Tcl_UnregisterChannel(backupMCB.interp, backupCB->inputChannel);
  Tcl_DeleteChannelHandler(backupCB->stderrChannel, 
			   ProcessBackupStderrChannel, backupCB);
  Tcl_UnregisterChannel(backupMCB.interp, backupCB->stderrChannel);
  Tcl_DeleteTimerHandler(backupCB->sentryTimer);
  
  /*
   * Call the cleanup procedure for this type, if one is registered
   */
  cleanupProc = Tcl_GetVar2(backupMCB.interp, "burt_cleanup_proc",
			    backupCB->packet.type, TCL_GLOBAL_ONLY);
  if (cleanupProc != NULL)
  {
    if (Tcl_VarEval(backupMCB.interp, cleanupProc, " ", 
		    backupCB->packet.host, " ", backupCB->packet.atom, 
		    " ", backupCB->packet.level, NULL) == TCL_ERROR)
    {
      sprintf(logBuffer, "BURT: Error detected in user defined cleanup "
	      "procedure %s: ", cleanupProc);
      Tcl_Write(backupMCB.logChannel, logBuffer, -1);
      Tcl_Write(backupMCB.logChannel, backupMCB.interp->result, -1);
      sprintf(logBuffer, " %s\n", BurtTime());
      Tcl_Write(backupMCB.logChannel, logBuffer, -1);
      Tcl_Flush(backupMCB.logChannel);
    }
  }

  /* Remove the backupCB entry from the backupCBTable */
  BurtSetupHashKey(key, backupCB->packet.host, backupCB->packet.atom,
		   backupCB->packet.level, backupCB->packet.type);
  entryPtr = Tcl_FindHashEntry(&backupCBTable, key);
  if (entryPtr != NULL)
    Tcl_DeleteHashEntry(entryPtr);

  /* 
   * If run state is STOP or ABORT, we don't want to start any more channels.
   */
    
  if ((BurtStatusGetRunState() & STATEMASK) != STOP && 
      (BurtStatusGetRunState() & STATEMASK) != ABORT)
  {
    done = 0;
    while (!done && BurtScheduleLength() > 0)
    {
      /*
       * If we get an error back when we try to start a channel, we should
       * try to start it again (with a different schedule item), until we have
       * either started the correct number of channels or run out of schedule 
       * items to try to start.
       */
      if (StartBackupChannel() == TCL_ERROR)
      {
	/* StartBackupChannel fills in an error, we just have to log it */
	Tcl_Write(backupMCB.logChannel, 
		  Tcl_GetStringResult(backupMCB.interp), -1);
	Tcl_Flush(backupMCB.logChannel);
	BurtStatusSetFinished(BurtStatusGetFinished() + 1);
      }
      else
      {
        done = 1;
      }
    }
  }

  /*
   * Check to see if all of the work is done, and set things to
   * reflect what we find.
   */
	
  if (BurtBackupIsComplete()) 
  {
    backupCB->packet.blockSize = END_OF_BACKUPSET;
    backupCB->packet.blockNum++;
    backupCB->packet.checksum = 0;
    /* AGG, must check for EOT here... */
    /* Write a filemark so we can seek to EOS (End Of Set) for appending */
    BurtMtWriteFilemark(backupMCB.outputChannel);
    BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetRunState(BACKUP | COMPLETE);
    BurtStatusSetCompleteTime(time((time_t *) NULL));
    Tcl_DeleteHashTable(&backupCBTable);
    Tcl_SetVar(backupMCB.interp, "burt_complete", "1", TCL_GLOBAL_ONLY);
  }
  free((void *) backupCB);
}


/*
 *----------------------------------------------------------------------
 *
 * EvaluateUserBackupProc --
 *
 *      Fill in the inputChannel and stderrChannel fields on the backup 
 *      control block by evaluating the user defined backup proc.  The
 *      backup proc must return two readable channels; the first is considered
 *      the input channel, along which the backup data is transmitted.  The
 *      second is the stderr channel, along which the standard error output
 *      is transmitted.
 *
 * Results:
 *	Returns TCL_OK if both channels were gathered, TCL_ERROR otherwise.
 *      If TCL_ERROR is returned, the interp->result field is filled
 *      in with an error message.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static int
EvaluateUserBackupProc(Tcl_Interp *interp, Burt_BackupControlBlock *backupCB)
{
  char *backupProc;
  int mode;
  int argc;
  int result;
  char **argv;
  
  /* Get the procedure name */
  backupProc = Tcl_GetVar2(interp, "burt_backup_proc", backupCB->packet.type, 
			   TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  if (backupProc == NULL) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "BURT: ERROR cannot find backup proc for type ",
		     backupCB->packet.type, " ", BurtTime(), "\n", NULL);
    return(TCL_ERROR);
  }

  /* Run the procedure */
  result = Tcl_VarEval(interp, backupProc, " ", 
		       backupCB->packet.host, " ",
		       backupCB->packet.atom, " ", 
		       backupCB->packet.level, 
		       NULL);
  if (result == TCL_ERROR)
  {
    FormatBackupProcError(interp, backupProc, 
			  backupCB->packet.host,
			  backupCB->packet.atom, 
			  backupCB->packet.level,
			  Tcl_GetStringResult(interp));
    return(TCL_ERROR);
  }

  /* User procedure should return a list of two channels */
  if (Tcl_SplitList(interp, interp->result, &argc, &argv) == TCL_ERROR) 
    return(TCL_ERROR);
  
  if (argc != 2) 
  {
    FormatBackupProcError(interp, backupProc, 
			  backupCB->packet.host, 
			  backupCB->packet.atom,
			  backupCB->packet.level, 
			  "Proc must return a two element list");
    return(TCL_ERROR);
  }
	
  /* Make sure the input channel exists and is readable */
  backupCB->inputChannel = Tcl_GetChannel(interp, argv[0], &mode);
  if (backupCB->inputChannel == NULL) 
  {
    FormatBackupProcError(interp, backupProc, 
			  backupCB->packet.host, 
			  backupCB->packet.atom,
			  backupCB->packet.level, 
			  "Cannot find input channel given by proc");
    return(TCL_ERROR);
  }

  if ((mode & TCL_READABLE) == 0) 
  {
    FormatBackupProcError(interp, backupProc, 
			  backupCB->packet.host, 
			  backupCB->packet.atom,
			  backupCB->packet.level, 
			  "Input channel given by proc is not readable");
    Tcl_UnregisterChannel(interp, backupCB->inputChannel);
    return(TCL_ERROR);
  }     
    
  /* The input channel should be non-blocking and in binary mode */
  result = Tcl_SetChannelOption(interp, 
				backupCB->inputChannel, 
				"-blocking", 
				"false");
  if (result != TCL_ERROR)
  {
    result = Tcl_SetChannelOption(interp,
				  backupCB->inputChannel,
				  "-translation",
				  "binary");
  }

  if (result == TCL_ERROR)
  {
    FormatBackupProcError(interp, backupProc,
			  backupCB->packet.host,
			  backupCB->packet.atom,
			  backupCB->packet.level,
			  Tcl_GetStringResult(interp));
    Tcl_UnregisterChannel(interp, backupCB->inputChannel);
    return(TCL_ERROR);
  }

  /* Make sure the stderr channel exists and is readable */
  backupCB->stderrChannel = Tcl_GetChannel(interp, argv[1], &mode);
  if (backupCB->stderrChannel == NULL) 
  {
    FormatBackupProcError(interp, backupProc, 
			  backupCB->packet.host, 
			  backupCB->packet.atom,
			  backupCB->packet.level, 
			  "Cannot find error channel given by proc");
    Tcl_UnregisterChannel(interp, backupCB->inputChannel);
    return(TCL_ERROR);
  }
  
  if ((mode & TCL_READABLE) == 0) 
  {
    FormatBackupProcError(interp, backupProc, 
			  backupCB->packet.host, 
			  backupCB->packet.atom,
			  backupCB->packet.level, 
			  "Error channel given by proc is not readable");
    Tcl_UnregisterChannel(interp, backupCB->inputChannel);
    Tcl_UnregisterChannel(interp, backupCB->stderrChannel);
    return(TCL_ERROR);
  }

  /* The error channel should be non-blocking */
  result = Tcl_SetChannelOption(interp, 
				backupCB->stderrChannel, 
				"-blocking", 
				"false");

  if (result == TCL_ERROR)
  {
    FormatBackupProcError(interp, backupProc,
			  backupCB->packet.host,
			  backupCB->packet.atom,
			  backupCB->packet.level,
			  Tcl_GetStringResult(interp));
    Tcl_UnregisterChannel(interp, backupCB->inputChannel);
    Tcl_UnregisterChannel(interp, backupCB->stderrChannel);
    return(TCL_ERROR);
  }

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessBackupInputChannel --
 *
 *	This procedure reads a block of data from the inputChannel of
 * the backup and arranges to get the data out onto the outputChannel.
 * It updates the packet header before writing the header, and data
 * block.  It is called from a Channel Handler, so there should be data 
 * waiting when it gets here.
 *
 * Side effects:
 *     May write data to the output stream
 *
 *----------------------------------------------------------------------
 */

static void  
ProcessBackupInputChannel(ClientData clientData, int mask)
{
  Burt_BackupControlBlock *backupCB;
  Burt_Packet EOTPacket;
  Tcl_Obj *resultPtr;
  static char buffer[1024];
  int nbytes;    /* Number of bytes read from inputChannel. */
  int mtbytes;   /* Number of bytes written to tape. */

  backupCB = (Burt_BackupControlBlock *) clientData;
  
  nbytes = Tcl_Read(backupCB->inputChannel,
		    &(backupCB->packet.buffer[backupCB->packet.blockSize]),
		    GetMaxPacketSize() - backupCB->packet.blockSize);
  
  if (nbytes > 0)
  {
    backupCB->packet.blockSize += nbytes;
    backupCB->packet.totalSize += nbytes;
  }
  else if (nbytes < 0) 
  {
    resultPtr = Tcl_GetObjResult(backupMCB.interp);
    Tcl_AppendStringsToObj(resultPtr, " ", BurtTime(), "\n", NULL);
    Tcl_Write(backupMCB.logChannel, Tcl_GetStringResult(backupMCB.interp), -1);
    Tcl_Flush(backupMCB.logChannel);
    /*
     * Kill this backup?  Yes.  Notify user, call it ?pipe error?
     */
    backupCB->packet.blockNum++;
    backupCB->packet.blockSize = ABORTED_BACKUP;
    backupCB->packet.checksum = 0; 
    
    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
    
    FlushBackupStderrChannel(backupCB);
    sprintf(buffer, "BURT: Aborted %s %s %s %s, pipe error %s\n",
	    backupCB->packet.host, backupCB->packet.atom,
	    backupCB->packet.level, backupCB->packet.type, BurtTime());
    Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
    Tcl_Flush(backupMCB.logChannel);
    
    CloseBackupChannel(backupCB);
    return;
  }

  if ((backupCB->packet.blockSize == GetMaxPacketSize() ||
      Tcl_Eof(backupCB->inputChannel)) && backupCB->packet.blockSize > 0)
  {
    backupCB->packet.blockNum++;
    backupCB->packet.checksum = BurtChecksum(backupCB->packet.buffer, 
					     backupCB->packet.blockSize);
    do
    {
      if (DEBUG)
      {
	struct tms foo;
	sprintf(buffer, "DEBUG: %s %s %s %s %lu\n", backupCB->packet.host,
		backupCB->packet.atom, backupCB->packet.level, 
		backupCB->packet.type, times(&foo));
	Tcl_Write(backupMCB.logChannel, buffer, -1);
	Tcl_Flush(backupMCB.logChannel);
      }

      if (backupCB->packet.blockNum == 1) 
      {
	BurtMtWriteFilemark(backupMCB.outputChannel);
      }

      mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
      switch (mtbytes)
      {
	case MT_EOT:
	case MT_ERROR:
	  {
	    /* Load a new tape */
	    if (mtbytes == MT_EOT)
	    {
	      Tcl_Write(backupMCB.logChannel, 
			"BURT: End of tape by MT_EOT ", -1);
	    }
	    else
	    {
	      Tcl_Write(backupMCB.logChannel, 
			"BURT: End of tape by MT_ERROR, ", -1);
	      Tcl_Write(backupMCB.logChannel, strerror(Tcl_GetErrno()), -1);
	      Tcl_Write(backupMCB.logChannel, " ", -1);
	    }
	    Tcl_Write(backupMCB.logChannel, BurtTime(), -1);
	    Tcl_Write(backupMCB.logChannel, "\n", -1);

	    /* Write a LOGICAL_EOT packet header */
	    EOTPacket.blockNum  = -1;
	    EOTPacket.blockSize = LOGICAL_EOT;
	    EOTPacket.totalSize = 0;
	    EOTPacket.host[0]   = '\0';
	    EOTPacket.atom[0]   = '\0';
	    EOTPacket.type[0]   = '\0';
	    EOTPacket.level[0]  = '\0';
	    EOTPacket.checksum  = 0;
	    
	    BurtMtWrite(backupMCB.outputChannel, &EOTPacket);
	
	    if (LoadNextBackupTape(backupMCB.outputChannel, 
				   backupMCB.logChannel) == TCL_ERROR) 
	    {
	      sprintf(buffer, "BURT: Backup error, unable to load tape %s\n",
		      BurtTime());
	      Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
	      Tcl_Flush(backupMCB.logChannel);
	      AbortBackup();
	      return;
	    }
	    break;
	  }
	case MT_FATALERROR:
	  {
	    /* Bad magic mojo going on, give up */
	    sprintf(buffer, "BURT: Backup fatal error, backups aborted %s\n", 
		    BurtTime());
	    Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
	    Tcl_Flush(backupMCB.logChannel);
	    AbortBackup();
	    return;
	    break;
	  }
	default:
	  {
	    /* Good fight, good write! */
	    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
	    /* reset the block size (ie, number of bytes collected so far) */
	    backupCB->packet.blockSize = 0;
	    break;
	  }
      }
    } while (mtbytes == MT_EOT || mtbytes == MT_ERROR);
  }

  if (Tcl_Eof(backupCB->inputChannel)) 
  {
    /* Write out a "End of Backup" packet */
    backupCB->packet.blockNum++;
    backupCB->packet.blockSize = END_OF_BACKUP;
    backupCB->packet.checksum = 0; 
    
    /* If this is the first packet, output a filemarker */
    if (backupCB->packet.blockNum == 1) 
    {
      /* If the first packet is also the last (ie, the channel disappeared
       * before sending any data, this is considered an aborted backup */
      backupCB->packet.blockSize = ABORTED_BACKUP;
      backupCB->packet.checksum = 0;
      BurtMtWriteFilemark(backupMCB.outputChannel);
    }

    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
    
    if (backupCB->packet.blockNum == 1) 
    {
      FlushBackupStderrChannel(backupCB);
      sprintf(buffer, "BURT: Aborted %s %s %s %s, 0 bytes and EOF on first "
	      "block of input %s\n", backupCB->packet.host, 
	      backupCB->packet.atom, backupCB->packet.level, 
	      backupCB->packet.type, BurtTime());
      Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
      Tcl_Flush(backupMCB.logChannel);
    } 
    else 
    {
      sprintf(buffer, 
	      "BURT: Completed %s %s %s %s, %d blocks, totaling %u bytes %s\n",
	      backupCB->packet.host, backupCB->packet.atom, 
	      backupCB->packet.level, backupCB->packet.type,
	      backupCB->packet.blockNum, backupCB->packet.totalSize, 
	      BurtTime());
      Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
      Tcl_Flush(backupMCB.logChannel);
    }
    CloseBackupChannel(backupCB);
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessBackupStderrChannel --
 *
 *	This procedure reads a stderr stream and re-writes the error to
 *      the logChannel with the dump stamp (I.E. the /@stall: identifier) 
 *      attached. It also calls the user defined procedure named in the
 *      backupMonitor field of the BackupControlBlock to check to see if
 *      the message was an error, or just plain old dialog.  If that proc
 *      returns anything but "0" the backup channels are closed down and
 *      a message regarding such action is placed in the log file.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void  
ProcessBackupStderrChannel(ClientData clientData, int mask)
{
  Burt_BackupControlBlock *backupCB;
  Tcl_DString buffer;
  Tcl_DString logMessage;
  Tcl_Obj *resultPtr;
  int nbytes;
  
  backupCB = (Burt_BackupControlBlock *) clientData;
  
  Tcl_DStringInit(&buffer);    
  nbytes = Tcl_Gets(backupCB->stderrChannel, &buffer);
  
  if (nbytes > 0) 
  {
    Tcl_DStringInit(&logMessage);    
    Tcl_DStringAppend(&logMessage, backupCB->packet.host, 
		      strlen(backupCB->packet.host));
    Tcl_DStringAppend(&logMessage, " ", -1);
    Tcl_DStringAppend(&logMessage, backupCB->packet.atom, 
		      strlen(backupCB->packet.atom));
    Tcl_DStringAppend(&logMessage, " ", -1);
    Tcl_DStringAppend(&logMessage, backupCB->packet.level, 
		      strlen(backupCB->packet.level));
    Tcl_DStringAppend(&logMessage, " ", -1);
    Tcl_DStringAppend(&logMessage, backupCB->packet.type, 
		      strlen(backupCB->packet.type));
    Tcl_DStringAppend(&logMessage, ": ", -1);
    Tcl_DStringAppend(&logMessage, Tcl_DStringValue(&buffer), nbytes);
    Tcl_DStringAppend(&logMessage, "\n", -1);
    Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&logMessage), 
	      Tcl_DStringLength(&logMessage));
    Tcl_Flush(backupMCB.logChannel);
    Tcl_DStringFree(&logMessage);
    
    /*
     * Call user defined proc to see if that log message should be
     * considered an error, or just plain dialog.
     */
    if (Tcl_VarEval(backupMCB.interp, backupCB->backupMonitor, " ", 
		    backupCB->packet.host, " ", backupCB->packet.atom, " ",
		    backupCB->packet.level, " ", 
		    "{", Tcl_DStringValue(&buffer), "}", NULL) == TCL_ERROR) {
      /* 
       * Tcl_VarEval found an error in user proc and filled in the error.
       */
      resultPtr = Tcl_GetObjResult(backupMCB.interp);
      Tcl_AppendStringsToObj(resultPtr, " ", BurtTime(), "\n", NULL);
      Tcl_Write(backupMCB.logChannel, 
		Tcl_GetStringResult(backupMCB.interp), -1);
      Tcl_Flush(backupMCB.logChannel);
    } 
    else 
    {
      if (strcmp(backupMCB.interp->result, "0") != 0) 
      {
	backupCB->packet.blockNum++;
	backupCB->packet.blockSize = ABORTED_BACKUP;
	backupCB->packet.checksum = 0; 
	      
	nbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
	BurtStatusSetBytes(BurtStatusGetBytes() + nbytes);
	
	FlushBackupStderrChannel(backupCB);
	Tcl_DStringInit(&logMessage); 
	Tcl_DStringAppend(&logMessage, "BURT: Aborted ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.host, 
			  strlen(backupCB->packet.host));
	Tcl_DStringAppend(&logMessage, " ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.atom, 
			  strlen(backupCB->packet.atom));
	Tcl_DStringAppend(&logMessage, " ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.level, 
			  strlen(backupCB->packet.level));
	Tcl_DStringAppend(&logMessage, " ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.type, 
			  strlen(backupCB->packet.type));
	Tcl_DStringAppend(&logMessage, 
			  ", user defined procedure ", -1);  
	Tcl_DStringAppend(&logMessage, backupCB->backupMonitor, 
			  strlen(backupCB->backupMonitor));
	Tcl_DStringAppend(&logMessage, " detected an error ", -1);
	Tcl_DStringAppend(&logMessage, BurtTime(), -1);
	Tcl_DStringAppend(&logMessage, "\n", -1);
	Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&logMessage), 
		  Tcl_DStringLength(&logMessage));
	Tcl_DStringFree(&logMessage);
	Tcl_Flush(backupMCB.logChannel);
	
	CloseBackupChannel(backupCB);
      }
    }
  }
  
  Tcl_DStringFree(&buffer);
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * FlushBackupStderrChannel --
 *
 *	This procedure flushes all available input on the stderr channel.
 *      It is meant to be used right before an atom is aborted, to ensure
 *      that all information from the stderr channel makes it into the log.
 *      Otherwise, when CloseBackupChannel is called, the errors are lost.
 *      It does no checking of the error messages, since the atom is already
 *      going to be aborted anyway.
 *
 *----------------------------------------------------------------------
 */
static void
FlushBackupStderrChannel(Burt_BackupControlBlock *backupCB)
{
  Tcl_DString buffer;
  Tcl_DString logMessage;
  int nbytes;
  
  while (Tcl_InputBuffered(backupCB->stderrChannel) > 0 && 
	 !Tcl_Eof(backupCB->stderrChannel)) 
  {
    Tcl_DStringInit(&buffer);    
    nbytes = Tcl_Gets(backupCB->stderrChannel, &buffer);
  
    if (nbytes > 0) 
    {
      Tcl_DStringInit(&logMessage);    
      Tcl_DStringAppend(&logMessage, backupCB->packet.host, 
			strlen(backupCB->packet.host));
      Tcl_DStringAppend(&logMessage, " ", -1);
      Tcl_DStringAppend(&logMessage, backupCB->packet.atom, 
			strlen(backupCB->packet.atom));
      Tcl_DStringAppend(&logMessage, " ", -1);
      Tcl_DStringAppend(&logMessage, backupCB->packet.level, 
			strlen(backupCB->packet.level));
      Tcl_DStringAppend(&logMessage, " ", -1);
      Tcl_DStringAppend(&logMessage, backupCB->packet.type, 
			strlen(backupCB->packet.type));
      Tcl_DStringAppend(&logMessage, ": ", -1);
      Tcl_DStringAppend(&logMessage, Tcl_DStringValue(&buffer), nbytes);
      Tcl_DStringAppend(&logMessage, "\n", -1);
      Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&logMessage), 
		Tcl_DStringLength(&logMessage));
      Tcl_Flush(backupMCB.logChannel);
      Tcl_DStringFree(&logMessage);
      Tcl_DStringFree(&buffer);
    }
    else
    {
      Tcl_DStringFree(&buffer);
      break;
    }
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessBackupSentry --
 *
 *	This procedure determines if a backup is stalled.  It is called
 *      every BURT_SENTRY_TIMEOUT seconds and confirms that data is moving 
 *      through the inputChannel.  If no data movement is found the backup 
 *      is considered stalled and is killed.
 *
 * Side effects:
 *      Normally just resets itself to be called in the future, but may
 *      kill the backup if it appears stalled.
 *----------------------------------------------------------------------
 */
static void 
ProcessBackupSentry(ClientData clientData)
{
  Burt_BackupControlBlock *backupCB;
  Tcl_DString buffer;
  int mtbytes;

  backupCB = (Burt_BackupControlBlock *) clientData;
  Tcl_DStringInit(&buffer);
  
  /* Two conditions must hold for a stream to be considered stalled:
   * 1) There is no data buffered on the stream; and
   * 2) There has been no data processed from this stream since the last
   *    sentry timeout
   */
  if ((Tcl_InputBuffered(backupCB->inputChannel) == 0) 
      && (backupCB->packet.totalSize == backupCB->sentry)) 
  {
    /* Stalled, close up the backup */
    backupCB->packet.blockNum++;
    backupCB->packet.blockSize = KILLED_BACKUP;
    backupCB->packet.checksum = 0; 
    
    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);

    Tcl_DStringAppend(&buffer, "BURT: Killed ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.host, -1);
    Tcl_DStringAppend(&buffer, " ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.atom, -1);
    Tcl_DStringAppend(&buffer, " ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.level, -1);
    Tcl_DStringAppend(&buffer, " ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.type, -1);
    Tcl_DStringAppend(&buffer, ", Sentry Timeout ", -1);
    Tcl_DStringAppend(&buffer, BurtTime(), -1);
    Tcl_DStringAppend(&buffer, "\n", -1);
    Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&buffer), 
	      Tcl_DStringLength(&buffer));
    
    CloseBackupChannel(backupCB);
  } 
  else 
  {
    /* 
     * Reset sentryTimer 
     */
    backupCB->sentry = backupCB->packet.totalSize;
    backupCB->sentryTimer = Tcl_CreateTimerHandler(BURT_SENTRY_TIMEOUT, 
						   ProcessBackupSentry, 
						   (ClientData) backupCB);
  }

  Tcl_DStringFree(&buffer);
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * BurtBackupIsComplete --
 *
 *	This procedure checks to see if there is any outstanding
 *      work to be done.
 *
 * Results:
 *      1 if there is nothing left to do, otherwise 0.
 *
 *----------------------------------------------------------------------
 */

static int
BurtBackupIsComplete(void)
{
  /*
   * There are three cases to check to see if Burt has completed 
   * its work:
   * 1) Run state = RUN, and burtBackupStatus.finished 
   *                             == burtBackupStatus.scheduled
   *    (I.E. the schedule Q is empty, and all the channels have finished)
   *
   * 2) Run state = STOP, and burtBackupStatus.finished 
   *                              == burtBackupStatus.started
   *    (I.E. the outstanding channels have finished up)
   * 3) Run state = COMPLETE (I.E., a backup has already run)
   */
  
  switch (BurtStatusGetRunState() & STATEMASK)
  {
    case RUN:
      {
	if (BurtStatusGetFinished() == BurtStatusGetScheduled())
	  return(1);
	break;
      }
    case STOP:
      {
	if (BurtStatusGetFinished() == BurtStatusGetStarted())
	  return(1);
	break;
      }
    case COMPLETE:
      {
	return(1);
	break;
      }
  }
  return(0);
}


/*
 *----------------------------------------------------------------------
 *
 * LoadNextBackupTape --
 *
 *	This procedure prompts the user (via the logChannel) to mount a
 *      new tape,   Before doing so it rewinds and unmounts the old one.  
 *      This procedure blocks until it gets a new tape and is able to write 
 *      a tape label. 
 *
 * Results:
 *      TCL_OK if the tape was swapped and a new tape label was
 *      written, TCL_ERROR otherwise.
 *      
 *----------------------------------------------------------------------
 */
static int
LoadNextBackupTape(Tcl_Channel mtChannel, Tcl_Channel logChannel)
{
  char buffer[1024];
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  Burt_BackupControlBlock *backupCB;
  Burt_TapeLabel expectedLabel, actualLabel;
  
  BurtStatusSetRunState(BACKUP | WAITING);

  sprintf(buffer, 
	  "BURT: End of tape %u in backup, wrote %1.2f megabytes %s\n",
	  BurtStatusGetTapes(), BurtStatusGetTotalXBytes(), BurtTime());
  Tcl_Write(logChannel, buffer, -1);

  /* Disable events on the backup channels */
  entryPtr = Tcl_FirstHashEntry(&backupCBTable, &searchPtr);
  while (entryPtr != NULL)
  {
    backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
    BurtDisableBackupEvents(backupCB);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }

  /* Make sure that the tape in the drive is what we expect it to be */
  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return(TCL_ERROR);

  if (BurtMtReadLabel(mtChannel, &actualLabel) != MT_OK)
    return(TCL_ERROR);

  SetupTapeLabel(&expectedLabel, 
		 BurtStatusGetTapeLabel(), 
		 BurtStatusGetTapes());
  if (TestTapeLabelEquality(&expectedLabel, &actualLabel) == 0)
  {
    sprintf(buffer, "BURT: ERROR, expected tape %s, #%s; got %s, #%s %s\n",
	    expectedLabel.label, 
	    expectedLabel.tape, 
	    actualLabel.label, 
	    actualLabel.tape, 
	    BurtTime());
    Tcl_Write(logChannel, buffer, -1);
    return(TCL_ERROR);
  }

  /* Readcheck this tape before ejecting it */
  savedBackupStatus = BurtStatusSave();
  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return(TCL_ERROR);

  if (BurtReadTape(backupMCB.interp, mtChannel, logChannel, 0) != TCL_OK)
  {
    /* BurtReadTape already filled in an error for us, we just write it out */
    Tcl_Write(logChannel, Tcl_GetStringResult(backupMCB.interp), -1);
    Tcl_Write(logChannel, "\n", -1);
    /* Make sure to restore the backup status */
    BurtStatusRestore(savedBackupStatus);
    BurtStatusFree(savedBackupStatus);
    return(TCL_ERROR);
  }

  /* Restore the saved backup status */
  BurtStatusRestore(savedBackupStatus);
  BurtStatusFree(savedBackupStatus);

  BurtStatusSetTapes(BurtStatusGetTapes() + 1);

  /* Phew!  Now we can load the next tape */
  sprintf(buffer, 
	  "BURT: Please mount tape %u in backup (%d seconds)...",
	  BurtStatusGetTapes(), TAPE_LOAD_DELAY);
  if (LoadNextTape(mtChannel, logChannel, buffer, TCL_WRITABLE) == TCL_ERROR)
  {
    Tcl_Write(logChannel, "BURT: ERROR in LoadNextBackupTape: \"", -1);
    Tcl_Write(logChannel, strerror(Tcl_GetErrno()), -1);
    Tcl_Write(logChannel, " ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
    return(TCL_ERROR);
  }
  
  sprintf(buffer, 
	  "BURT: Mounted tape %u in backup; thank you %s\n",
	  BurtStatusGetTapes(), BurtTime());
  Tcl_Write(logChannel, buffer, -1);
  BurtStatusSetRunState(BACKUP | RUN);

  /* Re-enable the events on the backup channels */
  entryPtr = Tcl_FirstHashEntry(&backupCBTable, &searchPtr);
  while (entryPtr != NULL)
  {
    backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
    BurtEnableBackupEvents(backupCB);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtBackupRunningList --
 *
 *      Creates a listing of all currently running atoms (ie, those that
 *      are active and are being written to tape NOW)
 *
 * Results:
 *	Returns a standard Tcl list of all the elements in backupCBTable. 
 *      Null if the table is empty.  The caller is responsible for the 
 *      freeing of memory.  NOTE: It returns this list in a Tcl_DString *
 *      not just a char *!
 *      
 *----------------------------------------------------------------------
 */

Tcl_DString *
BurtBackupRunningList(void)
{
  Tcl_DString *runningList;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Burt_BackupControlBlock *backupCB;
    
  runningList = (Tcl_DString *)malloc(sizeof(Tcl_DString));
  if (runningList != NULL) 
  {
    Tcl_DStringInit(runningList);
    if ((BurtStatusGetRunState() & STATEMASK) != ABORT &&
	(BurtStatusGetRunState() & STATEMASK) != COMPLETE)
    {
      entryPtr = Tcl_FirstHashEntry(&backupCBTable, &search);
      while (entryPtr != NULL) 
      {
	backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
	Tcl_DStringStartSublist(runningList);
	Tcl_DStringAppendElement(runningList, backupCB->packet.host);
	Tcl_DStringAppendElement(runningList, backupCB->packet.atom);
	Tcl_DStringAppendElement(runningList, backupCB->packet.level);
	Tcl_DStringAppendElement(runningList, backupCB->packet.type);
	Tcl_DStringEndSublist(runningList);
	entryPtr = Tcl_NextHashEntry(&search);
      }
    }
  }

  return(runningList);
}

/*
 *----------------------------------------------------------------------
 *
 * FormatBackupProcError
 *
 *      Formats an error message caused by a problem with a user-defined
 *      backup proc
 *
 * Side effects:
 *      Overwrites interp->result with a new error message.
 *      
 *----------------------------------------------------------------------
 */
void 
FormatBackupProcError(Tcl_Interp *interp, char *procName, 
		      char *host, char *atom, char *level, char *errorMessage)
{
  /* Use a Tcl_Obj to build up the error message, then make it the interp
   * result at the end.  That way, the function can handle the case where
   * the errorMessage is actually in interp->result, and this function was
   * called like this: FormatBackupProcError(..., Tcl_GetStringResult(interp))
   */
  Tcl_Obj *resultPtr;
  resultPtr = Tcl_NewStringObj("", -1);
  Tcl_AppendStringsToObj(resultPtr,
			 "BURT: ERROR in backup proc, procedure:  ", 
			 procName, "\n",
			 "BURT: ERROR in backup proc, parameters: ", 
			 host, " ", atom, " ", level, "\n",
			 "BURT: ERROR in backup proc, message:    ",
			 errorMessage, "\n",
			 "BURT: ERROR in backup proc, time:       ",
			 BurtTime(), "\n", NULL);
  Tcl_SetObjResult(interp, resultPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtEnableBackupEvents
 *
 *      Enables backup events on the channels in a backup control block
 *
 *----------------------------------------------------------------------
 */
void
BurtEnableBackupEvents(Burt_BackupControlBlock *backupCB)
{
  Tcl_Channel channelName;
  int channelMode;
  Tcl_ChannelProc *channelProc;
  ClientData clientData;
  int delay;
  Tcl_TimerProc *timerProc;

  /* Set up the channel handler for the backup input channel */
  channelName = backupCB->inputChannel;
  channelMode = TCL_READABLE;
  channelProc = ProcessBackupInputChannel;
  clientData  = (ClientData)backupCB;
  Tcl_CreateChannelHandler(channelName, channelMode, channelProc, clientData);

  /* Set up the channel handler for the stderr channel */
  channelName = backupCB->stderrChannel;
  channelMode = TCL_READABLE;
  channelProc = ProcessBackupStderrChannel;
  clientData  = (ClientData)backupCB;
  Tcl_CreateChannelHandler(channelName, channelMode, channelProc, clientData);

  /* Set the initial sentry timeout value */
  backupCB->sentry = -1;

  /* Set up the sentry timer handler */
  delay      = BURT_SENTRY_TIMEOUT;
  timerProc  = ProcessBackupSentry;
  clientData = (ClientData)backupCB;
  backupCB->sentryTimer = Tcl_CreateTimerHandler(delay, timerProc, clientData);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtDisableBackupEvents
 *
 *      Disable backup events on the channels in a backup control block
 *
 *----------------------------------------------------------------------
 */
void
BurtDisableBackupEvents(Burt_BackupControlBlock *backupCB)
{
  Tcl_Channel channelName;
  int channelMode;
  Tcl_ChannelProc *channelProc;
  ClientData clientData;

  /* Disable events on the input channel by setting the event mode to 0 */
  channelName = backupCB->inputChannel;
  channelMode = 0;
  channelProc = ProcessBackupInputChannel;
  clientData  = (ClientData)backupCB;
  Tcl_CreateChannelHandler(channelName, channelMode, channelProc, clientData);

  /* Disable events on the stderr channel by setting the event mode to 0 */
  channelName = backupCB->stderrChannel;
  channelMode = 0;
  channelProc = ProcessBackupStderrChannel;
  clientData  = (ClientData)backupCB;
  Tcl_CreateChannelHandler(channelName, channelMode, channelProc, clientData);
  
  /* Disable the sentry timer by deleting the timer event */
  Tcl_DeleteTimerHandler(backupCB->sentryTimer);
}


/*
 *----------------------------------------------------------------------
 *
 * GetUserBackupMonitorProc
 *
 *      Find the name of the backupmonitor proc for a particular backup
 *      control block.  This information comes from the global Tcl array
 *      burt_backupmonitor_proc, which is keyed by type.
 *
 * Results:
 *      TCL_OK if a backupmonitor proc was registered for the type given in
 *      the backupCB; TCL_ERROR otherwise, and interp->result is set to an
 *      error message.
 *      
 * Side effects:
 *      Fills in the backupMonitor field of the backupCB
 *      
 *----------------------------------------------------------------------
 */
static int
GetUserBackupMonitorProc(Tcl_Interp *interp, Burt_BackupControlBlock *backupCB)
{
  backupCB->backupMonitor = Tcl_GetVar2(interp, 
					"burt_backupmonitor_proc",
					backupCB->packet.type,
					TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  
  if (backupCB->backupMonitor == NULL) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, 
		     "BURT: ERROR cannot find backupmonitor proc for type ",
		     backupCB->packet.type, " ", BurtTime(), "\n", NULL);
    return(TCL_ERROR);
  }

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * GetBackupVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetBackupVersion(void)
{
  return(rcsid);
}
