/*
 * burtChecksum.c
 *
 *    This file implements a simple Fletcher checksum for burt.
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the software.
 * A reasonable copying fee may be charged for any distribution of this 
 * software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no representations
 * about the suitability of this software for any purpose.  It is provided 
 * "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#define _POSIX_SOURCE
#ifndef lint
static char rcsid[] = "$Id: burtChecksum.c,v 1.5 1999/04/01 20:54:29 jmelski Exp $";
#endif

#include <burtInt.h> 



/*
 *----------------------------------------------------------------------
 *
 * BurtChecksum --
 *
 *	This procedure calculates a simple checksum of a buffer. 
 *      Fletcher checksum:
 *      The basic formula.
 *
 *      check0 = check0 + b;
 *      check1 = check1 + check0;
 *
 *      1. You should use unsigned char for b.  (Otherwise, you may be 
 *         subtracting rather than checksumming.)
 *
 *      2. You are right in that the sum won't overflow a 
 *         32 bit unsigned int.  In fact, many of the higher level bits 
 *         will not be sent.  For this reason, why not return  
 *         (c1 & 0xffff) << 16 + (c0 & 0xffff)?
 *         This would provide more significant bits in the checksum.
 *         NOTE: We assume that the buffer is no larger than BURT_MAX_BLOCKSIZE,
 *         due to the fact that that is the largest buffer we can
 *         currently write to a tape.
 *
 * Results:
 *	
 *	Returns the checksum.
 *
 * Side effects:
 * 
 *----------------------------------------------------------------------
 */

unsigned int 
BurtChecksum(const void *buffer, int bufferLen)
{
    unsigned char *byteBuffer;  /* View this buffer a byte at a time. */
    int check0;        /* The first check sum for the final checksum. */
    int check1;        /* The second check sum for the final cheksum. */
    int i;

    byteBuffer = (unsigned char *)buffer;
    check0 = 0;
    check1 = 0;

    for (i = 0; i < bufferLen; i++) {
	check0 += (int)byteBuffer[i];
	check1 += check0;
    }
    
    return(((check1 & 0xffff) << 16) + (check0 & 0xffff));
}


/*
 *----------------------------------------------------------------------
 *
 * GetChecksumVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetChecksumVersion(void)
{
  return(rcsid);
}
