/*
 * burtInt.h
 *
 *    This file defines constants and functions used internally by Burt
 *
 * Copyright (C) 1995 - 1999 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: burtInt.h,v 1.28 1999/04/07 23:43:57 jmelski Exp $
 */

#ifndef _BURTINT
#define _BURTINT  
#define _POSIX_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>

#include <tcl.h>
#include <tclInt.h>
#include <burt.h>

#ifndef EXIT_FAILURE 
	#define EXIT_FAILURE 1
#endif
#ifndef EXIT_SUCCESS 
	#define EXIT_SUCCESS 0
#endif


/*
 * Time delays
 */
#define BURT_SENTRY_TIMEOUT 300000 /* (5 Min) Time between sentry checks in */
				   /* milliseconds, this keeps track of */
				   /* backup progress. */
#define TAPE_LOAD_DELAY 60         /* Number of seconds to wait on each loop */
                                   /* of the tape loading routines */


#define MAX_IO_RETRIES 5           /* Maximum times to retry rewind/eject */

/*
 * Run States.
 */

#define NONE         1
#define BACKUP       2
#define READTAPE     4
#define RECOVER      8

#define RUN          0
#define STOP        16
#define COMPLETE    32
#define ABORT       64
#define WAITING    128

#define OPMASK      15
#define STATEMASK  240
  
/*
 * Special Packet Sizes.
 */

#define END_OF_BACKUP       -1    /* End of an atom backup */
#define END_OF_BACKUPSET    -2    /* End of a burt process (ie, no more tapes) */
                                  /* Burt puts one of the tape when */
                                  /* it finishes running */
#define LOGICAL_EOT         -3    /* Logical EOT packet */
#define ABORTED_BACKUP      -4    /* End of an aborted atom backup */
#define KILLED_BACKUP       -5    /* End of a killed atom backup */


/*
 * Mag Tape (MT) interface return codes.
 */
#define MT_OK          0     /* Tape request OK. */
#define MT_ERROR       -1    /* Tape error. */
#define MT_EOT         -2    /* Physical End of Tape. */
#define MT_VEOT        -3    /* Virtual End of Tape.  This can happen */
                             /* _anywhere_. */
#define MT_FATALERROR  -4    /* Unrecoverable tape error */
#define xBYTES         1048576  /* Multiplier for tape size */

typedef struct Burt_ScheduleItem {
    char host[BURT_MAX_HOSTLEN]; /* Host on which atom lives. */
    char atom[BURT_MAX_ATOMLEN]; /* The filesys/volume (/usr, u.jmelski) */
    char type[BURT_MAX_TYPELEN]; /* User defined dump type - ufs, afs... */
    char level[BURT_MAX_LEVELLEN];  /* The level of this dump. */
} Burt_ScheduleItem;

extern char *GetMainVersion(void);

extern int Burt_BackupObjCmd(ClientData cdata, Tcl_Interp *interp, 
			     int objc, Tcl_Obj *CONST objv[]);

extern int Burt_RecoverCmd(ClientData cdata, Tcl_Interp *interp,
			  int argc, char **argv);

extern int Burt_ScheduleObjCmd(ClientData cdata, Tcl_Interp *interp, 
			       int objc, Tcl_Obj *CONST objv[]);

extern int Burt_ReadTapeCmd(ClientData cdata, Tcl_Interp *interp,
			    int argc, char **argv);

extern int Burt_StatusCmd(ClientData cdata, Tcl_Interp *interp, 
			  int argc, char **argv);

extern int Burt_MtObjCmd(ClientData cdata, Tcl_Interp *interp, int objc, 
			 Tcl_Obj *CONST objv[]);

/* begin misc functions defined in burtBackup.c */
extern Tcl_DString *BurtBackupRunningList(void);
extern char *GetBackupVersion(void);
/* end misc functions defined in burtBackup.c */

/* begin misc functions defined in burtRecover.c */
extern Tcl_DString *BurtRecoverRunningList(void);
extern char *GetRecoverVersion(void);
/* end misc functions defined in burtRecover.c */

/* begin misc functions defined in burtReadtape.c */
Tcl_DString *BurtReadtapeRunningList(void);
extern void BurtReadtapeInit(void);
extern int BurtReadTape(Tcl_Interp *interp, Tcl_Channel inputChannel, 
			Tcl_Channel logChannel, int readMultipleTapes);
extern char *GetReadtapeVersion(void);
/* end misc functions defined in burtReadtape.c */

/*
 * begin "public" functions from the schedule module 
 */
extern void BurtScheduleInit(void);
extern int BurtScheduleAdd(Tcl_Interp *interp, char *host, 
			   char *atom, char *level, char *type);
extern int BurtScheduleRemove(Tcl_Interp *interp, char *host, 
			      char *atom, char *level, char *type);
extern Burt_ScheduleItem *BurtScheduleGetNext(Tcl_Interp *interp, 
					      Tcl_Channel logChannel);
extern Tcl_DString *BurtScheduleList(void);
extern char *BurtScheduleInfo(void);
extern int BurtScheduleLength(void);
extern char *GetScheduleVersion(void);
/*
 * end "public" functions from the schedule module 
 */

extern void BurtStatusReset(void);
extern char *BurtStatusSave(void);
extern void BurtStatusRestore(char *restoreStatus);
extern void BurtStatusFree(char *freeStatus);
extern void BurtStatusSetRunState(int runState);
extern int BurtStatusGetRunState(void);
extern void BurtStatusSetTapeLabel(char *);
extern char *BurtStatusGetTapeLabel(void);
extern void BurtStatusSetScheduled(unsigned int numScheduled);
extern unsigned int BurtStatusGetScheduled(void);
extern void BurtStatusSetChannels(unsigned int numChannels);
extern unsigned int BurtStatusGetChannels(void);
extern void BurtStatusSetFinished(unsigned int numFininshed);
extern unsigned int BurtStatusGetFinished(void);
extern void BurtStatusSetStarted(unsigned int numStarted);
extern unsigned int BurtStatusGetStarted(void);
extern void BurtStatusSetTapes(unsigned int numTapes);
extern unsigned int BurtStatusGetTapes(void);
extern void BurtStatusSetBytes(unsigned int numBytes);
extern unsigned int BurtStatusGetBytes(void);
extern void BurtStatusSetStartTime(time_t startTime);
extern time_t BurtStatusGetStartTime(void);
extern void BurtStatusSetCompleteTime(time_t startTime);
extern time_t BurtStatusGetCompleteTime(void);
extern void BurtStatusSetXBytes(unsigned int numXBytes);
extern unsigned int BurtStatusGetXBytes(void);
extern void BurtStatusSetSnapBytes(unsigned int numBytes);
extern unsigned int BurtStatusGetSnapBytes(void);
extern void BurtStatusSetSnapXBytes(unsigned int numXBytes);
extern unsigned int BurtStatusGetSnapXBytes(void);
extern void BurtStatusSetSnapTime(time_t snapTime);
extern time_t BurtStatusGetSnapTime(void);
extern float BurtStatusGetCurrentRate(time_t now);
extern float BurtStatusGetRate(time_t now);
extern float BurtStatusGetTotalXBytes(void);
extern void GetRuntimeStatus(Tcl_Interp *interp);
extern char *GetStatusVersion(void);

extern unsigned int BurtChecksum(const void *buffer, int bufferLen);
extern char *GetChecksumVersion(void);

extern int BurtMtWrite(Tcl_Channel mtChannel, Burt_Packet *packet);
extern int BurtMtRead(Tcl_Channel mtChannel, Burt_Packet *packet);
extern int BurtMtWriteLabel(Tcl_Channel mtChannel, Burt_TapeLabel *label);
extern int BurtMtReadLabel(Tcl_Channel mtChannel, Burt_TapeLabel *returnLabel);
extern int BurtMtWriteFilemark(Tcl_Channel mtChannel);
extern int BurtMtFastForward(Tcl_Channel mtChannel, int count);
extern int BurtMtOffline(Tcl_Channel mtChannel);
extern int BurtMtStatus(Tcl_Channel mtChannel);
extern int BurtMtEOM(Tcl_Channel mtChannel);
extern int BurtMtBSF(Tcl_Channel mtChannel, int count);
extern int BurtMtRewind(Tcl_Channel mtChannel);
extern int BurtMtSafeRewind(Tcl_Channel mtChannel);
extern int BurtMtSafeOffline(Tcl_Channel mtChannel);
extern char *GetMtIOVersion(void);

extern char *BurtTime(void);
extern int  BurtTableLength(Tcl_HashTable *tablePtr);
extern void BurtSetupHashKey(char *key, char *host, char *atom, 
			     char *level, char *type);
extern int  LoadNextTape(Tcl_Channel mtChannel, Tcl_Channel logChannel, 
			 char *prompt, int mode);
extern void SetupTapeLabel(Burt_TapeLabel *tapeLabel, 
			   char *label, 
			   unsigned int tape);
extern int  TestTapeLabelEquality(Burt_TapeLabel *labelA, 
				  Burt_TapeLabel *labelB);
extern char *GetUtilVersion(void);
extern int  GetMaxPacketSize(void);
#endif /* _BURTINT */     

