/*
 * burtMtIO.c
 *
 *    This file implements the mag tape interface for burt.
 *
 * Copyright (C) 1995 - 1999 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifndef lint
static char rcsid[] = "$Id: burtMtIO.c,v 1.23 1999/04/07 23:43:58 jmelski Exp $";
#endif

#ifdef _POSIX_SOURCE
    #undef _POSIX_SOURCE
#endif
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>  
#include <netinet/in.h>
#include <signal.h>
#define _POSIX_SOURCE
#include <unistd.h>
#include <math.h>
#include <burtInt.h> 


static void HeaderToNetworkOrder(Burt_Packet *packetPtr);
static void HeaderToHostOrder(Burt_Packet *packetPtr);

/*
 *----------------------------------------------------------------------
 *
 * Burt_MtObjCmd --
 *
 *	This procedure implements the Tcl command mt; it is a "transaction
 *      center" -- based on the input, it calls one of several functions to
 *      do some task.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.  
 *
 * Side effects:
 *      May do something to the tape, depending on the particular function.
 *	
 *----------------------------------------------------------------------
 */

int 
Burt_MtObjCmd(ClientData cdata, Tcl_Interp *interp, 
	      int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Channel mtChannel;
  int mode;
  int index;
  static char *options[] = { "offline", "rewind", "status", NULL };
  enum options { MT_OFFLINE, MT_REWIND, MT_STATUS };

  if (objc != 3)
  {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
    return(TCL_ERROR);
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0, 
			  &index) != TCL_OK)
    return(TCL_ERROR);

  mtChannel = 
    Tcl_GetChannel(interp, Tcl_GetStringFromObj(objv[2], NULL), &mode);

  if (mtChannel == NULL)
    return TCL_ERROR;

  switch ((enum options) index)
  {
    case MT_OFFLINE:
    {
      if (BurtMtOffline(mtChannel) == MT_ERROR)
      {
	Tcl_SetResult(interp, "Unable to offline channel", TCL_STATIC);
	return(TCL_ERROR);
      }
      break;
    }
    case MT_REWIND:
    {
      if (BurtMtRewind(mtChannel) == MT_ERROR)
      {
	Tcl_SetResult(interp, "Unable to rewind channel", TCL_STATIC);
	return(TCL_ERROR);
      }
      break;
    }
    case MT_STATUS:
    {
      if (BurtMtStatus(mtChannel) == MT_ERROR)
      {
	Tcl_SetResult(interp, "Unable to set channel status", TCL_STATIC);
	return(TCL_ERROR);
      }
      break;
    }
  }

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtMtWrite --
 *
 *	This procedure writes a packet containing header information and
 *      the backup data to tape.
 *
 * Results:
 *	
 *	Returns the number of bytes written (the size of the packet) if 
 *      all went well, otherwise MT_EOT upon end of tape, or MT_ERROR 
 *      if there was a write error.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int
BurtMtWrite(Tcl_Channel mtChannel, Burt_Packet *packetPtr)
{
  int fd;                /* Unix level file descriptor. */
  int nbytes;
  
  Tcl_GetChannelHandle(mtChannel, TCL_WRITABLE, (ClientData *) &fd);
  HeaderToNetworkOrder(packetPtr);
  nbytes = write(fd, (void *) packetPtr, sizeof(Burt_Packet));
  HeaderToHostOrder(packetPtr);
  
  /* 
   * Possible returns from write()          Significance
   *   0                                    MT_EOT
   *   error                                MT_ERROR
   *   short
   *     short, 0                           MT_EOT
   *     short, error                       MT_FATALERROR
   */

  if (nbytes == 0) 
    return (MT_EOT);
  else if (nbytes < 0)
    return (MT_ERROR);
  
  if (nbytes != sizeof(Burt_Packet)) 
  {
    HeaderToNetworkOrder(packetPtr);
    nbytes = write(fd, (void *) packetPtr, sizeof(Burt_Packet));
    HeaderToHostOrder(packetPtr);
    if (nbytes == 0)
      return(MT_EOT);
    else 
      return(MT_FATALERROR);
  }

  return(nbytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtRead --
 *
 *	This procedure reads a packet from tape.  It deals with all the details
 *      of doing so, including converting to host byte order.
 *
 * Results:
 *	
 *	Returns the number of bytes read if all went well; otherwise, it
 *      returns an error code:
 *             MT_EOT          End of tape (physical OR logical)
 *             MT_ERROR        A recoverable error
 *             MT_FATALERROR   An unrecoverable error
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int 
BurtMtRead(Tcl_Channel mtChannel, Burt_Packet *packetPtr)
{
  int fd;             /* Unix level file descriptor. */
  int nbytes;

  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);

  /* Read possibilities:
   *
   * return value               significance
   *   sizeof(Burt_Packet)        read a packet
   *   0                          read EOF
   *     0, sizeof(Burt_Packet)     EOF, followed by packet
   *     0, 0                       MT_EOT
   *     0, error                   ???, probably MT_FATALERROR
   *     0, short                   MT_ERROR
   *   short                      MT_ERROR
   *   error                      MT_FATALERROR
   */

  nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
  if (nbytes == 0)
  {
    /* Read a file marker (or EOF); re-read to get packet */
    nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
    if (nbytes == 0) 
    {
      return(MT_EOT);
    } 
    else if (nbytes != sizeof(Burt_Packet))
    {
      return(MT_ERROR);
    }
    else if (nbytes < 0) 
    {
      return(MT_FATALERROR);
    }
  } 
  else if (nbytes != sizeof(Burt_Packet)) 
  {
    return(MT_ERROR);
  }
  else if (nbytes < 0)
  {
    return(MT_FATALERROR);
  }

  HeaderToHostOrder(packetPtr);

  if (packetPtr->blockSize == LOGICAL_EOT)
    return(MT_EOT);

  return(nbytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtWriteLabel --
 *
 *	This procedure writes a tape label onto the tape.
 *      It should be used on each new tape in a backup. 
 *
 * Results:
 *	
 *      Number of bytes written if the label was written, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtWriteLabel(Tcl_Channel mtChannel, Burt_TapeLabel *label)
{
  int fd;              /* Unix level file descriptor. */
  int nbytes;
	   
  Tcl_GetChannelHandle(mtChannel, TCL_WRITABLE, (ClientData *) &fd);

  /* 
   * Ick!  The MtIO layer shouldn't be doing this, but the user layer
   * shouldn't know that it needs to be done!  Argh...
   */
  label->magic = (int)htonl(label->magic);

  nbytes = write(fd, (void *) label, sizeof(Burt_TapeLabel));
  
  if (nbytes != sizeof(Burt_TapeLabel))
    return(MT_ERROR);
  return(nbytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtReadLabel --
 *
 *	This procedure reads a tape label from the tape.  It copies the
 *      label data to returnLabel, unless 
 *      returnLabel is NULL.
 *
 * Results:
 *	
 *      MT_OK if the label was read, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int
BurtMtReadLabel(Tcl_Channel mtChannel, Burt_TapeLabel *returnLabel)
{
  int fd;              /* Unix level file descriptor. */
  Burt_TapeLabel label;
  int nbytes;
  
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);

  nbytes = read(fd, (void *) &label, sizeof(Burt_TapeLabel));
  
  if (nbytes != sizeof(Burt_TapeLabel)) 
  {
    /* See if the label is the size of the old tape labels */
    if (nbytes != 72)
    {
      returnLabel = NULL;
      return(MT_ERROR);
    }
  }

  if (returnLabel != NULL)
  {
    memcpy((void *)returnLabel, (void *)&label, sizeof(Burt_TapeLabel));
    /* 
     * Ick!  The MtIO layer shouldn't be doing this, but the user layer
     * shouldn't know that it needs to be done!  Argh...
     */
    returnLabel->magic = (int)ntohl(returnLabel->magic);
  }

  return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtFastForward --
 *
 *	This procedure reads "count" filemarks on the tape to "fast forward"
 *      over data.
 *
 * Results:
 *	
 *      MT_OK if the filemarks were read, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtFastForward(Tcl_Channel mtChannel, int count)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTFSF;
    mt_command.mt_count = count;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtMtWriteFilemark --
 *
 *	This procedure writes a filemark onto the tape.
 *
 * Results:
 *	
 *      MT_OK if the filemark was written, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtWriteFilemark(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_WRITABLE, (ClientData *) &fd);
    mt_command.mt_op = MTWEOF;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtMtRewind --
 *
 *	This procedure rewinds tape.
 *
 * Results:
 *	
 *      MT_OK if the tape was rewound, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtRewind(Tcl_Channel mtChannel)
{
  int fd;              /* Unix level file descriptor. */
  struct mtop mt_command;
  Tcl_Seek(mtChannel, 0, SEEK_SET);    /* "Rewind" files */
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
  mt_command.mt_op = MTREW;
  mt_command.mt_count = 1;
  if (ioctl(fd, MTIOCTOP, &mt_command) < 0)
    return (MT_ERROR);

  return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtSafeRewind --
 *
 *      Wrapper around BurtMtRewind that handles multiple retries
 *      
 * Results:
 *      
 *      MT_ERROR if all the retries fail; MT_OK otherwise.
 *	
 * Side effects:
 *      
 *      Rewinds the tape
 *      
 *----------------------------------------------------------------------
 */

int
BurtMtSafeRewind(Tcl_Channel mtChannel)
{
  int retries = 0;
  while (retries < MAX_IO_RETRIES && BurtMtRewind(mtChannel) == MT_ERROR)
    retries++;

  if (retries == MAX_IO_RETRIES)
    return (MT_ERROR);

  return (MT_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtMtOffline --
 *
 *	This procedure takes a tape offline.
 *
 * Results:
 *	
 *      MT_OK if the tape was taken offline, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtOffline(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTOFFL;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0)
      return(MT_ERROR);

    return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtSafeOffline --
 *
 *      Wrapper around BurtMtOffline that handles multiple retries
 *      
 * Results:
 *      
 *      MT_ERROR if all the retries fail; MT_OK otherwise.
 *	
 * Side effects:
 *      
 *      Offlines the tape
 *      
 *----------------------------------------------------------------------
 */

int BurtMtSafeOffline(Tcl_Channel mtChannel)
{
  int retries = 0;
  while (retries < MAX_IO_RETRIES && BurtMtOffline(mtChannel) == MT_ERROR)
    retries++;

  if (retries == MAX_IO_RETRIES)
    return (MT_ERROR);

  return (MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtStatus --
 *
 *	This procedure sets the status of a tape drive.
 *
 * Results:
 *	
 *      MT_OK if the tape status was set, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtStatus(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTNOP;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtEOM --
 *
 *	This procedure advances the tape to the end of media marker
 *
 * Results:
 *	
 *      MT_OK if the tape status was set, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtEOM(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTEOM;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtBSF --
 *
 *	This procedure rewinds the tape count filemarks
 *
 * Results:
 *	
 *      MT_OK if the tape was rewound, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtBSF(Tcl_Channel mtChannel, int count)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTBSF;
    mt_command.mt_count = count;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * HeaderToNetworkOrder --
 *
 *	This procedure converts a packet header from host byte order
 *      to network byte order.
 *
 * Results:
 *	
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static void 
HeaderToNetworkOrder(Burt_Packet *packetPtr)
{
    packetPtr->magic = (int)htonl(packetPtr->magic);
    packetPtr->blockNum = (int) htonl(packetPtr->blockNum);
    packetPtr->blockSize = (int) htonl(packetPtr->blockSize);
    packetPtr->totalSize = (unsigned int) htonl(packetPtr->totalSize);
    packetPtr->checksum = (unsigned int) htonl(packetPtr->checksum);

    return;
}


/*
 *----------------------------------------------------------------------
 *
 * HeaderToHostOrder --
 *
 *	This procedure converts a packet header from network byte order
 *      to the host byte order.
 *
 * Results:
 *	
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static void 
HeaderToHostOrder(Burt_Packet *packetPtr)
{
    packetPtr->magic = (int)ntohl(packetPtr->magic);
    packetPtr->blockNum = (int) ntohl(packetPtr->blockNum);
    packetPtr->blockSize = (int) ntohl(packetPtr->blockSize);
    packetPtr->totalSize = (unsigned int) ntohl(packetPtr->totalSize);
    packetPtr->checksum = (unsigned int) ntohl(packetPtr->checksum);

    return;
}


/*
 *----------------------------------------------------------------------
 *
 * GetMtIOVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetMtIOVersion(void)
{
  return(rcsid);
}
