/*
 * burtReadtape.c
 *
 *    This file implements the readtape command.  
 *
 * Copyright (C) 1995 - 1999 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef lint
static char rcsid[] = "$Id: burtReadtape.c,v 1.31 1999/04/15 18:42:23 jmelski Exp $";
#endif

#include <burtInt.h>

#define DEBUG 0

typedef struct Burt_TOCItem {
  char key[BURT_MAX_HOSTLEN+
	  BURT_MAX_ATOMLEN+
	  BURT_MAX_LEVELLEN+
	  BURT_MAX_TYPELEN+1];
  unsigned int tape;
  unsigned int offset;
} Burt_TOCItem;

static Tcl_HashTable burtTOC;
static Tcl_HashTable burtInitTOC;
static int startedReadtape;
static char *savedReadtapeStatus;
static char *BurtGetTapeTOC(void);
static int LoadNextReadTape(Tcl_Channel mtChannel, Tcl_Channel logChannel);

/*
 *----------------------------------------------------------------------
 *
 * Burt_ReadtapeCmd --
 *
 *	This procedure implements the Tcl command readtape.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

int 
Burt_ReadtapeObjCmd(ClientData cdata, Tcl_Interp *interp, 
		    int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Channel inputChannel, logChannel;
  char *inputChannelName, *logChannelName;
  int index, mode;
  static char *options[] = { "label", "toc", NULL };
  enum options { READTAPE_LABEL, READTAPE_TOC };

  if (objc != 4)
  {
    Tcl_WrongNumArgs(interp, 1, objv, "option inputchannel logchannel");
    return(TCL_ERROR);
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], options, 
			  "option", 0, &index) != TCL_OK)
    return(TCL_ERROR);

  /* Get the names of the input and log channels */
  inputChannelName = Tcl_GetStringFromObj(objv[2], NULL);
  logChannelName = Tcl_GetStringFromObj(objv[3], NULL);

  /* Make sure that the input channel is a channel, and that it's readable */
  inputChannel = Tcl_GetChannel(interp, inputChannelName, &mode);
  if (inputChannel == NULL)
    return(TCL_ERROR);

  if ((mode & TCL_READABLE) == 0)
  {
    Tcl_SetResult(interp, "input channel wasn't opened for reading", 
		  TCL_STATIC);
    return(TCL_ERROR);
  }

  logChannel = Tcl_GetChannel(interp, logChannelName, &mode);
  if (logChannel == NULL)
    return(TCL_ERROR);

  if ((mode & TCL_WRITABLE) == 0)
  {
    Tcl_SetResult(interp, "log channel wasn't opened for writing", TCL_STATIC);
    return(TCL_ERROR);
  }
  
  if (BurtMtRewind(inputChannel) != MT_OK)
  {
    Tcl_Write(logChannel, "BURT: WARNING, unable to rewind tape ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
  }
  return(TCL_OK);
}


int 
Burt_ReadTapeCmd(ClientData cdata, Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_Channel inputChannel;
  Tcl_Channel logChannel;
  char *TOC;
  Burt_TapeLabel label;
  int mode;
  char logBuffer[1024];

  if (argc != 4) 
  {
    Tcl_SetResult(interp, "wrong # args: should be \"readtape option "
		  "inputchannel logchannel\"", TCL_STATIC);
    return(TCL_ERROR);
  }

  if (!((strcmp(argv[1], "toc") == 0) || (strcmp(argv[1], "label") == 0))) 
  {
    Tcl_AppendResult(interp, "bad option \"", argv[1],
		     "\": should be toc or label", (char *) NULL);
    return(TCL_ERROR);
  }

  inputChannel = Tcl_GetChannel(interp, argv[2], &mode);
  if (inputChannel == NULL) 
    /* Tcl_GetChannel fills in the error message for us. */
    return(TCL_ERROR);

  if ((mode & TCL_READABLE) == 0) 
  {
    Tcl_SetResult(interp, "input channel wasn't opened for reading", 
		  TCL_STATIC);
    return(TCL_ERROR);
  }     

  logChannel = Tcl_GetChannel(interp, argv[3], &mode);
  if (logChannel == NULL)
    /* Tcl_GetChannel fills in the error message for us. */
    return(TCL_ERROR);

  if ((mode & TCL_WRITABLE) == 0) 
  {
    Tcl_SetResult(interp, "log channel wasn't opened for writing", TCL_STATIC);
    return(TCL_ERROR);
  }     

  if (BurtMtRewind(inputChannel) == MT_ERROR) 
  {
    Tcl_Write(logChannel, "BURT: WARNING unable to rewind tape ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
  }

    
  BurtStatusReset();

  if (strcmp(argv[1], "toc") == 0) 
  {
    if (BurtReadTape(interp, inputChannel, logChannel, 1) == TCL_OK) {
      TOC = BurtGetTapeTOC();
      if (TOC == NULL) 
      {
	Tcl_SetResult(interp, "BURT: Could not allocate memory for TOC!\n", 
		      TCL_STATIC);
	BurtStatusSetRunState(READTAPE | ABORT);
	BurtStatusSetCompleteTime(time((time_t *) NULL));
	BurtReadtapeInit();
	return(TCL_ERROR);
      } 
      else 
      {
	Tcl_SetResult(interp, TOC, TCL_DYNAMIC);
	BurtStatusSetRunState(READTAPE | COMPLETE);
	BurtStatusSetCompleteTime(time((time_t *) NULL));
	BurtReadtapeInit();
	return(TCL_OK);
      }
    } 
    else 
    {
      sprintf(logBuffer, "BURT: Outputting partial TOC, %1.2f bytes, "
	      "%d salvagable items\n", BurtStatusGetTotalXBytes(), 
	      BurtTableLength(&burtTOC));
      Tcl_AppendResult(interp, logBuffer, (char *)NULL);
      TOC = BurtGetTapeTOC();
      if (TOC == NULL)
      {
	Tcl_AppendResult(interp, "BURT: Could not allocate memory for TOC!\n",
			 (char *)NULL);
      }
      else
      {
	Tcl_AppendResult(interp, TOC, (char *)NULL);
	free((void *)TOC);
      }

      BurtStatusSetRunState(READTAPE | ABORT);
      BurtStatusSetCompleteTime(time((time_t *) NULL));
      return(TCL_ERROR);
    }
  } 
  else if (strcmp(argv[1], "label") == 0)
  {
    if (BurtMtReadLabel(inputChannel, &label) != MT_OK)
    {
      Tcl_SetResult(interp, "Could not read tape label", TCL_STATIC);
      return(TCL_ERROR);
    }

    sprintf(logBuffer, "%s %s %d {%s}", label.label, label.tape, label.magic, 
	    label.date);
	    
    Tcl_AppendResult(interp, logBuffer, (char *) NULL);
    BurtStatusSetRunState(READTAPE | COMPLETE);
    BurtStatusSetCompleteTime(time((time_t *) NULL));
    return(TCL_OK);
  }
  return(TCL_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtReadtapeInit --
 *
 *	This procedure simply sets the startedReadtape flag to zero.  The
 *      flag is used by readtape to determine if partial readchecks have 
 *      already been done.
 *
 * Side effects:
 *      Sets startedReadtape to zero.
 *      
 *----------------------------------------------------------------------
 */

void
BurtReadtapeInit(void)
{
  startedReadtape = 0;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtReadTape --
 *
 *	This procedure reads a burt tape and validates the data contained
 *      on it.  In the process it builds a tape table of contents, which
 *      may be retrieved by calling BurtGetTapeTOC().
 *
 * Results:
 *     TCL_OK if tape is OK, TCL_ERROR otherwise, interp-> result is
 *     filled in with the error if TCL_ERROR
 *
 * Side effects:
 *     Creates a TOC for the tape.
 *      
 *----------------------------------------------------------------------
 */

int 
BurtReadTape(Tcl_Interp *interp, Tcl_Channel inputChannel, 
	     Tcl_Channel logChannel, int readMultipleTapes)
{
  Burt_Packet packet;
  Burt_TOCItem *TOCItemPtr, *tempTOCItemPtr;
  Tcl_HashEntry *entryPtr;
  int mtStatus;
  int new;
  int tapeOffset = 0;
  char logBuffer[1024];
  Burt_TapeLabel label;
  
  if (BurtMtReadLabel(inputChannel, &label) != MT_OK)
  {
    Tcl_SetResult(interp, "Could not read tape label", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* If we've already started readchecks, reload the status info for them */
  if (startedReadtape == 1)
  {
    BurtStatusRestore(savedReadtapeStatus);
    BurtStatusFree(savedReadtapeStatus);
    
    /* This tape should belong to the same set as "previously viewed" tapes */
    if (strcmp(BurtStatusGetTapeLabel(), label.label) != 0)
    {
      Tcl_SetResult(interp, "Current tape is not from the correct backup set.",
		    TCL_STATIC);
      return(TCL_ERROR);
    }
  }
  else  /* startedReadtape == 0 */
  {
    Tcl_InitHashTable(&burtInitTOC, TCL_STRING_KEYS);
    Tcl_InitHashTable(&burtTOC, TCL_STRING_KEYS);
    
    BurtStatusReset();
    BurtStatusSetStartTime(time((time_t *) NULL));
  }

  BurtStatusSetRunState(READTAPE | RUN);
  BurtStatusSetTapeLabel(label.label);
  BurtStatusSetChannels(1);
  BurtStatusSetTapes((atoi(label.tape)));
  startedReadtape = 1;
  
  TOCItemPtr = (Burt_TOCItem *) malloc(sizeof(Burt_TOCItem));

  do 
  {
    /* Let the Tcl event loop process some events */
    while (Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT) != 0);

    mtStatus = BurtMtRead(inputChannel, &packet);
    switch (mtStatus)
    {
      case MT_EOT:
      case MT_ERROR:
	/* Both MT_EOT and MT_ERROR cause us to load a new tape */
	{
	  if (mtStatus == MT_EOT)
	  {
	    Tcl_Write(logChannel, "BURT: End of tape by MT_EOT ", -1);
	  }
	  else
	  {
	    Tcl_Write(logChannel, "BURT: End of tape by MT_ERROR, ", -1);
	  }

	  Tcl_Write(logChannel, BurtTime(), -1);
	  Tcl_Write(logChannel, "\n", -1);
	  
	  if (readMultipleTapes == 0)
	  {
	    savedReadtapeStatus = BurtStatusSave();
	    free((void *)TOCItemPtr);
	    return (TCL_OK);
	  }

	  if (LoadNextReadTape(inputChannel, logChannel) == TCL_ERROR) 
	  {
	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, 
			     "BURT: --------------------------------------\n",
			     "BURT: Readtape error, unable to load tape\n",
			     "BURT: --------------------------------------\n",
			     (char *) NULL);
	    free ((void *)TOCItemPtr);
	    return (TCL_ERROR);
	  }
	  continue;
	  break;
	}
      case MT_FATALERROR:
	{
	  /* MT_FATALERROR is some bad mojo going on.  Best quit. */
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp, 
			   "BURT: ------------------------------------\n",
			   "BURT: Readtape fatal error, readchecks fail\n",
			   "BURT: ", strerror(Tcl_GetErrno()), "\n",
			   "BURT: ------------------------------------\n",
			   (char *) NULL);
	  free ((void *)TOCItemPtr);
	  return(TCL_ERROR);
	  break;
	}
    }

    if (packet.blockSize != END_OF_BACKUPSET)
    {
      if ((BurtChecksum((void *) packet.buffer, 
			packet.blockSize) != packet.checksum)) 
      {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, 
			 "BURT: ------------------------\n",
			 "BURT: Readtape Checksum Error: ", 
			 packet.host, " ", 
			 packet.atom, " ", 
			 packet.level, " ", 
			 packet.type, " ",
			 BurtTime(), "\n"
			 "BURT: ------------------------\n",
			 (char *) NULL);
	free ((void *)TOCItemPtr);
	return(TCL_ERROR);
      }
      else /* checksums match */
      {
	/* add this atom to the TOC if this is a first block */
	if (packet.blockNum == 1) 
	{
	  tapeOffset++;
	  strcpy(TOCItemPtr->key, packet.host);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.atom);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.level);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.type);
	  TOCItemPtr->tape = BurtStatusGetTapes();
	  TOCItemPtr->offset = tapeOffset;
	  if (DEBUG) Tcl_Write(logChannel, "filemarker\n", -1);
	  
	  entryPtr = Tcl_CreateHashEntry(&burtInitTOC, TOCItemPtr->key, &new);
	  
	  if (new) 
	  {
	    Tcl_SetHashValue(entryPtr, TOCItemPtr);
	    TOCItemPtr = (Burt_TOCItem *) malloc(sizeof(Burt_TOCItem));
	    BurtStatusSetStarted(BurtStatusGetStarted() + 1);
	    sprintf(logBuffer, "BURT: Created entry {{%s %s %s %s} %d %d} "
		    "%s\n",
		    packet.host, packet.atom, packet.level, packet.type,
		    BurtStatusGetTapes(), tapeOffset, BurtTime());
	    Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(logChannel);
	  } 
	}
	
	/* If this is the end of a backup, put out a message */
	if (packet.blockSize == END_OF_BACKUP)
	{
	  strcpy(TOCItemPtr->key, packet.host);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.atom);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.level);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.type);
	  
	  BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	  
	  entryPtr = Tcl_FindHashEntry(&burtInitTOC, TOCItemPtr->key);
	  if (entryPtr != NULL)
	  {
	    tempTOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
	    Tcl_DeleteHashEntry(entryPtr);
	    
	    entryPtr = Tcl_CreateHashEntry(&burtTOC, tempTOCItemPtr->key, &new);
	    if (new)
	      Tcl_SetHashValue(entryPtr, tempTOCItemPtr);

	    sprintf(logBuffer, "BURT: Completed entry {{%s %s %s %s} %d %d} "
		    "%s\n", packet.host, packet.atom, packet.level, 
		    packet.type, tempTOCItemPtr->tape, tempTOCItemPtr->offset,
		    BurtTime());
	  }
	  else
	  {
	    sprintf(logBuffer, "BURT: Completed entry {{%s %s %s %s} ? ?} "
		    "%s\n", packet.host, packet.atom, packet.level, 
		    packet.type, BurtTime());
	  }
	  Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	  Tcl_Flush(logChannel);
	}

	/* if this is an aborted or killed dump, remove it from the TOC */
	if (packet.blockSize == ABORTED_BACKUP || 
	    packet.blockSize == KILLED_BACKUP)
	{
	  strcpy(TOCItemPtr->key, packet.host);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.atom);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.level);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.type);
	  
	  entryPtr = Tcl_FindHashEntry(&burtInitTOC, TOCItemPtr->key);
	  if (entryPtr != NULL)
	  {
	    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	    tempTOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
	    sprintf(logBuffer, "BURT: Removed entry {{%s %s %s %s} %d %d} "
		    "%s\n",
		    packet.host, packet.atom, packet.level, packet.type,
		    tempTOCItemPtr->tape, tempTOCItemPtr->offset, BurtTime());
	    Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(logChannel);
	    free((void *)tempTOCItemPtr);
	    Tcl_DeleteHashEntry(entryPtr);
	  }
	}
      }
    }
    BurtStatusSetBytes(BurtStatusGetBytes() + mtStatus);
  }
  while (packet.blockSize != END_OF_BACKUPSET);
    
  free((void *)TOCItemPtr);
  BurtStatusSetRunState(READTAPE | COMPLETE);
  BurtStatusSetCompleteTime(time((time_t *) NULL));
  return(TCL_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtGetTapeTOC --
 *
 *     This procedure returns the tape's TOC (as built by calling 
 *     BurtReadTape).  It is the callers responsibility to free the TOC 
 *     list it returns.
 *
 * Results:
 *      The tape's Table of Contents (if there was one created) or NULL.    
 *     
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static char *
BurtGetTapeTOC(void)
{
  Tcl_DString TOCList;
  Burt_TOCItem *TOCItemPtr;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  char buffer[64];
  char *TOCStr;

  Tcl_DStringInit(&TOCList);
  entryPtr = Tcl_FirstHashEntry(&burtTOC, &search);
  while (entryPtr != NULL) 
  {
    TOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
    Tcl_DStringStartSublist(&TOCList);
    Tcl_DStringAppendElement(&TOCList, TOCItemPtr->key);
    sprintf(buffer, "%d", TOCItemPtr->tape);
    Tcl_DStringAppendElement(&TOCList, buffer); 
    sprintf(buffer, "%d", TOCItemPtr->offset);
    Tcl_DStringAppendElement(&TOCList, buffer); 
    Tcl_DStringEndSublist(&TOCList);
    entryPtr = Tcl_NextHashEntry(&search);
  }

  TOCStr = (char *)malloc(Tcl_DStringLength(&TOCList) + 11);
  sprintf(TOCStr, "BURT: TOC %s", Tcl_DStringValue(&TOCList));
  
  Tcl_DStringFree(&TOCList);
  return(TOCStr);
}

/*
 *----------------------------------------------------------------------
 *
 * LoadNextReadTape --
 *
 *	Arrange for a new tape to be loaded, via LoadNextTape()
 *
 * Results:
 *      TCL_OK if the tape was swapped successfully; TCL_ERROR otherwise
 *
 * Side effects:
 *      Loads a new tape, via LoadNextTape
 *      
 *----------------------------------------------------------------------
 */

static int
LoadNextReadTape(Tcl_Channel mtChannel, Tcl_Channel logChannel)
{
  char buffer[1024];
  
  BurtStatusSetRunState(READTAPE | WAITING);

  sprintf(buffer, 
	  "BURT: End of tape %u in readtape, read %1.2f megabytes %s\n",
	  BurtStatusGetTapes(), BurtStatusGetTotalXBytes(), BurtTime());
  Tcl_Write(logChannel, buffer, -1);

  BurtStatusSetTapes(BurtStatusGetTapes() + 1);
  
  sprintf(buffer, 
	  "BURT: Please mount tape %u in readtape (%d seconds)...",
	  BurtStatusGetTapes(), TAPE_LOAD_DELAY);
  if (LoadNextTape(mtChannel, logChannel, buffer, TCL_READABLE) == TCL_ERROR)
  {
    Tcl_Write(logChannel, "BURT: ERROR in LoadNextReadTape: \"", -1);
    Tcl_Write(logChannel, strerror(Tcl_GetErrno()), -1);
    Tcl_Write(logChannel, " ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
    return(TCL_ERROR);
  }
  
  sprintf(buffer, 
	  "BURT: Mounted tape %u in readtape; thank you %s\n",
	  BurtStatusGetTapes(), BurtTime());
  Tcl_Write(logChannel, buffer, -1);
  BurtStatusSetRunState(READTAPE | RUN);
  return(TCL_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtReadtapeRunningList --
 *
 *      Creates a listing of all currently running atoms (ie, those that
 *      are active and are being written to tape NOW)
 *
 * Results:
 *	Returns a standard Tcl list of all the elements in backupCBTable. 
 *      Null if the table is empty.  The caller is responsible for the 
 *      freeing of memory.  NOTE: It returns this list in a Tcl_DString *
 *      not just a char *!
 *      
 *----------------------------------------------------------------------
 */

Tcl_DString *
BurtReadtapeRunningList(void)
{
  Tcl_DString *runningList;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Burt_TOCItem *TOCItemPtr;
  
  runningList = (Tcl_DString *)malloc(sizeof(Tcl_DString));
  if (runningList != NULL) 
  {
    Tcl_DStringInit(runningList);
    entryPtr = Tcl_FirstHashEntry(&burtInitTOC, &search);
    while (entryPtr != NULL) 
    {
      TOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
      Tcl_DStringAppendElement(runningList, TOCItemPtr->key);
      entryPtr = Tcl_NextHashEntry(&search);
    }
  }

  return(runningList);
}


/*
 *----------------------------------------------------------------------
 *
 * GetReadtapeVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetReadtapeVersion(void)
{
  return(rcsid);
}
