/*
 * burtSchedule.c
 *
 *    This file implements the schedule module.  It allows schedule items
 *    to be added, deleted, listed, etc...  This is where the Tcl command
 *    "schedule" is found.
 *
 * Copyright (C) 1995 - 1999 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  
 * It is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#define _POSIX_SOURCE
#ifndef lint
static char rcsid[] = "$Id: burtSchedule.c,v 1.18 1999/04/01 20:54:31 jmelski Exp $";
#endif

#include <burtInt.h>
/* Declare functions accessible only to the schedule module */
static int BurtScheduleCheckInputLength(Tcl_Interp *interp, char *host, 
					char *atom,  char *level, char *type);

/* Declare data accessible only to the schedule module */
static Tcl_HashTable scheduleTable;   /* Table of scheduled dumps */


/*
 *----------------------------------------------------------------------
 *
 * Burt_ScheduleObjCmd --
 *
 *	This procedure implements the Tcl command schedule.  It is a 
 *      "transaction center" -- based on the particular input, it calls
 *      one of several functions, relaying a portion of the input, to 
 *      do a task.
 *
 * Results:
 *	Returns a standard Tcl completion code, based on the result of
 *      the daughter functions.
 *
 * Side effects:
 *       scheduleTable may be modified (by one of the daughter functions)
 *
 *----------------------------------------------------------------------
 */

int
Burt_ScheduleObjCmd(ClientData cdata, Tcl_Interp *interp, 
		    int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Obj *resultPtr;
  Tcl_DString *scheduleList;
  int index;
  char *host, *atom, *level, *type;
  static char *options[] = { "add", "info", "length", 
			     "list", "remove", NULL };
  enum options { SCHEDULE_ADD, SCHEDULE_INFO, SCHEDULE_LENGTH, 
		 SCHEDULE_LIST, SCHEDULE_REMOVE };

  if (objc < 2)
  {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
    return(TCL_ERROR);
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0, 
			  &index) != TCL_OK)
    return(TCL_ERROR);

  switch ((enum options) index)
  {
    case SCHEDULE_ADD:
    case SCHEDULE_REMOVE:
    {
      if (objc != 6) 
      {
	Tcl_WrongNumArgs(interp, 2, objv, "host atom level type");
	return(TCL_ERROR);
      }
      
      host = Tcl_GetStringFromObj(objv[2], NULL);
      atom = Tcl_GetStringFromObj(objv[3], NULL);
      level = Tcl_GetStringFromObj(objv[4], NULL);
      type = Tcl_GetStringFromObj(objv[5], NULL);
      
      if (index == SCHEDULE_ADD)
      {
	if (BurtScheduleAdd(interp, host, atom, level, type) != TCL_OK)
	  return(TCL_ERROR);
      }
      else
      {
	if (BurtScheduleRemove(interp, host, atom, level, type) != TCL_OK)
	  return(TCL_ERROR);
      }
      return(TCL_OK);
      break;
    }
    case SCHEDULE_LENGTH:
    {
      if (objc != 2)
      {
	Tcl_WrongNumArgs(interp, 2, objv, NULL);
	return(TCL_ERROR);
      }
      
      resultPtr = Tcl_GetObjResult(interp);
      Tcl_SetIntObj(resultPtr, BurtScheduleLength());
      return(TCL_OK);
      break;
    }
    case SCHEDULE_LIST:
    {
      scheduleList = BurtScheduleList();
      if (scheduleList == NULL)
      {
	Tcl_SetResult(interp, "Could not allocate memory for list", 
		      TCL_STATIC);
	return(TCL_ERROR);
      }
      else
      {
	Tcl_DStringResult(interp, scheduleList);
	Tcl_DStringFree(scheduleList);
	Tcl_Free((char *)scheduleList);
	return(TCL_OK);
      }
      break;
    }    
    case SCHEDULE_INFO:
    {
      if (objc != 2)
      {
	Tcl_WrongNumArgs(interp, 2, objv, NULL);
	return(TCL_ERROR);
      }
      
      Tcl_SetResult(interp, BurtScheduleInfo(), TCL_DYNAMIC);
      return(TCL_OK);
      break;
    }
  }
  return(TCL_OK);
}
		    

/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleInit --
 *
 *	This procedure sets up the schedule table.
 *
 * Results:
 *	None
 *
 * Side effects:
 *       *
 *----------------------------------------------------------------------
 */

void
BurtScheduleInit(void)
{
  /* No need to check that the table is successfully initialized, because
   * the Tcl_InitHashTable function *can't* fail -- it is guaranteed to 
   * either work or cause the interpreter to abort
   */
    Tcl_InitHashTable(&scheduleTable, TCL_STRING_KEYS);
    return;
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleCheckInputLength --
 *
 *	This procedure ensures that the input to the schedule command is
 *      of a suitable length (ie, that the length of the host, atom, level
 *      and type are < the length of BURT_MAX_HOSTLEN, BURT_MAX_ATOMLEN,
 *      BURT_MAX_LEVELLEN, BURT_MAX_TYPELEN, respectively)
 *
 * Results:
 *	Returns a standard Tcl completion code, sets interpreter result on
 *      error.
 *
 * Side effects:
 *       None
 *
 *----------------------------------------------------------------------
 */

int
BurtScheduleCheckInputLength (Tcl_Interp *interp, 
			      char *host, char *atom, char *level, char *type)
{
  static char errorMessage[128];
  char parts[4][6] = { "host", "atom", "level", "type" };
  char *specifier[4];
  int maxes[4] = { BURT_MAX_HOSTLEN, BURT_MAX_ATOMLEN, 
		   BURT_MAX_LEVELLEN, BURT_MAX_TYPELEN };
  int size;
  int i;

  specifier[0] = host;
  specifier[1] = atom;
  specifier[2] = level;
  specifier[3] = type;

  /* Verify that the length of each part is suitably short */
  for (i = 0; i < 4; i++)
  {
    size = strlen(specifier[i]);
    if (size >= maxes[i])
    {
      /* Set the Tcl interpreter result and return error */
      sprintf(errorMessage, "Invalid item: %s is too long (%d chars, "
	      "should be less than %d chars)", parts[i], size, maxes[i]);
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, errorMessage, (char *) NULL);
      return(TCL_ERROR);
    }
  }

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleAdd --
 *
 *	This procedure adds a Burt_ScheduleItem struct to the schedule 
 *      table.
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *       scheduleTable grows, Burt_ScheduleItem is allocated
 *
 *----------------------------------------------------------------------
 */

int
BurtScheduleAdd (Tcl_Interp *interp, 
		 char *host, char *atom, char *level, char *type)
{
  Tcl_HashEntry *entryPtr;
  Burt_ScheduleItem *scheduleItemPtr;
  int new;
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];

  /* First make sure that the input isn't too long */
  if (BurtScheduleCheckInputLength(interp, host, atom, level, type) != TCL_OK)
    return(TCL_ERROR);

  /* Pack the host, atom, level and type data into a hash key */
  BurtSetupHashKey(key, host, atom, level, type);

  entryPtr = Tcl_CreateHashEntry(&scheduleTable, key, &new);
  
  if (new) 
  {
    /* Stuff args into an Burt_ScheduleItem struct and enqueue it. */
    scheduleItemPtr = 
      (Burt_ScheduleItem *)Tcl_Alloc(sizeof(Burt_ScheduleItem));
    strncpy(scheduleItemPtr->host, host, BURT_MAX_HOSTLEN);
    strncpy(scheduleItemPtr->atom, atom, BURT_MAX_ATOMLEN);
    strncpy(scheduleItemPtr->level, level, BURT_MAX_LEVELLEN);
    strncpy(scheduleItemPtr->type, type, BURT_MAX_TYPELEN);
  
    Tcl_SetHashValue(entryPtr, scheduleItemPtr);
  } 
  else 
  {
    /* There is already an entry in the table for this schedule item. */
    /* Maybe this should be an error, but I don't really think so -- who 
     * cares if you try to multiply insert something?  It's like calling the
     * "set" command repeatedly.  It's harmless.
     *
     * return(TCL_ERROR);
     */
  }
  
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleRemove --
 *
 *	This procedure removes a Burt_ScheduleData struct from the
 *      schedule table. It does not return it, rather it throws it
 *      away.
 *
 * Results:
 *	Returns a standard Tcl completion code; on error, sets interpreter
 *      result to an error description
 *
 * Side effects:
 *       scheduleTable shrinks.
 *
 *----------------------------------------------------------------------
 */

int
BurtScheduleRemove (Tcl_Interp *interp,
		    char *host, char *atom, char *level, char *type)
{
  Tcl_HashEntry *entryPtr;
  Burt_ScheduleItem *scheduleItem;
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  /* Result must be big enough for "Cannot to find "host atom level type"" */
  char resultString[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
		   BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 32];
  
  /* First check that the input isn't too long */
  if (BurtScheduleCheckInputLength(interp, host, atom, level, type) != TCL_OK)
    return(TCL_ERROR);

  /* Pack the host, atom, level and type into a hash key */
  BurtSetupHashKey(key, host, atom, level, type);
  
  entryPtr = Tcl_FindHashEntry(&scheduleTable, key);
  if (entryPtr != NULL) 
  {
    scheduleItem = (Burt_ScheduleItem *)Tcl_GetHashValue(entryPtr);
    Tcl_DeleteHashEntry(entryPtr);
    Tcl_Free((char *)scheduleItem);
  } 
  else 
  {
    sprintf(resultString, "Cannot find \"%s %s %s %s\"", 
	    host, atom, level, type);
    Tcl_SetResult(interp, resultString, TCL_STATIC);
    return(TCL_ERROR);
  }

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleGetNext --
 *
 *	This procedure returns the next scheduled item from the table.
 *      It also removes thie entry in the schedule table.
 *
 * Results:
 *	Returns the item to caller, it is the callers responsibility to
 *      free the storage.  NULL is returned if the table is empty.
 *
 * Side effects:
 *      scheduleTable shrinks.
 *----------------------------------------------------------------------
 */

Burt_ScheduleItem *
BurtScheduleGetNext(Tcl_Interp *interp, Tcl_Channel logChannel)
{
  Burt_ScheduleItem *scheduleItemPtr;
  Tcl_HashEntry *entryPtr, *entryFoo;
  Tcl_HashSearch search;
  
  char *orderingProc;
  int argc;
  char **argv;
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  
  /* The first entry in the hash table is the default next item */
  entryPtr = Tcl_FirstHashEntry(&scheduleTable, &search);
  
  /* See if there's a user-defined ordering procedure.  If so, call it
   * to determine which item to start next.  If the return from that 
   * procedure is bad in any way, give up, print an error, use the default
   */
  orderingProc = Tcl_GetVar(interp, "burt_ordering_proc", TCL_GLOBAL_ONLY);
  if (orderingProc != NULL)
  {
    if (Tcl_VarEval(interp, orderingProc, NULL) == TCL_ERROR)
    {
      Tcl_Write(logChannel, "BURT: Error detected in user defined "
		"ordering proc.", -1);
    } 
    else
    {
      if (Tcl_SplitList(interp, Tcl_GetStringResult(interp), 
			&argc, &argv) == TCL_ERROR)
      {
	Tcl_Write(logChannel, "BURT: Error while parsing return from user "
		  "defined ordering proc.", -1);
      }
      else
      {
	if (argc != 4)
	{
	  if (argc > 0)
	    Tcl_Write(logChannel, "BURT: Error in user defined ordering "
		      "proc; must return a list of four elements.", -1);
	}
	else 
	{
	  if (BurtScheduleCheckInputLength(interp, argv[0], argv[1],
					   argv[2], argv[3]) != TCL_OK)
	  {
	    Tcl_Write(logChannel, "BURT: User defined ordering proc "
		      "returned an invalid item: ", -1);
	    Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
	    Tcl_Write(logChannel, "\n", -1);
	  }
	  else
	  {
	    BurtSetupHashKey(key, argv[0], argv[1], argv[2], argv[3]);
	    entryFoo = Tcl_FindHashEntry(&scheduleTable, key);
	    Tcl_Free((char *)argv);
	    if (entryFoo != NULL) 
	      entryPtr = entryFoo;
	  }
	}
      }
    }
  }
  
  if (entryPtr == NULL)
    return(NULL);
  
  scheduleItemPtr = (Burt_ScheduleItem *)Tcl_GetHashValue(entryPtr);
  Tcl_DeleteHashEntry(entryPtr);
  return(scheduleItemPtr);
}

/*
Tcl_HashEntry *
BurtScheduleGetUserDefinedNext(Tcl_Interp *interp, Tcl_Channel logChannel)
{
  Tcl_HashEntry *entryPtr;
  char *procName;
  int argc;
  char **argv;
  int result;
  char key[BURT_MAX_HOSTLEN + 
	  BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + 
	  BURT_MAX_TYPELEN + 1];

  procName = Tcl_GetVar(interp, "burt_ordering_proc", TCL_GLOBAL_ONLY);
  if (procName == NULL)
    return(NULL);

  result = Tcl_VarEval(interp, procName, NULL);
  if (result == TCL_ERROR)
  {
    FormatUserOrderProcError(interp, procName, Tcl_GetStringResult(interp));
    Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
    return(NULL);
  } 

  result = Tcl_SplitList(interp, Tcl_GetStringResult(interp), &argc, &argv);
  if (result == TCL_ERROR)
  {
    FormatUserOrderProcError(interp, procName, "Unable to parse return value");
    Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
    return(NULL);
  }

  if (argc != 4)
  {
    if (argc > 0)
    {
      FormatUserOrderProcError(interp, 
			       procName, 
			       "Return value must be a four element list");
      Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
    }
    return(NULL);
  }

  result = BurtScheduleCheckInputLength(interp, 
					argv[0], 
					argv[1], 
					argv[2], 
					argv[3]);
  if (result != TCL_OK)
  {
    FormatUserOrderProcError(interp, procName, "Return value is invalid");
    Tcl_Write(logChannel, Tcl_GetStringResult(interp), -1);
    return(NULL);
  }

  BurtSetupHashKey(key, argv[0], argv[1], argv[2], argv[3]);
  entryPtr = Tcl_FindHashEntry(&scheduleTable, key);
  Tcl_Free((char *)argv);
  if (entryPtr == NULL) 
    return(NULL);

  return(entryPtr);
}
*/

/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleList --
 *
 *	This procedure creates a listing of all the scheduled items in
 *      the schedule table.
 *
 * Results:
 *	Returns a standard Tcl list of all the elements in the schedule table.
 *      Null if the table is empty.  The caller is responsible for the 
 *      freeing of memory.  NOTE: It returns this list in a Tcl_DString *
 *      not just a char *!
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

Tcl_DString *
BurtScheduleList(void)
{
  Tcl_DString *scheduleList;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Burt_ScheduleItem *scheduleItemPtr;
  
  scheduleList = (Tcl_DString *)Tcl_Alloc(sizeof(Tcl_DString));
  if (scheduleList != NULL) 
  {
    Tcl_DStringInit(scheduleList);
    entryPtr = Tcl_FirstHashEntry(&scheduleTable, &search);
    while (entryPtr != NULL) 
    {
      scheduleItemPtr = (Burt_ScheduleItem *)Tcl_GetHashValue(entryPtr);
      Tcl_DStringStartSublist(scheduleList);
      Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->host);
      Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->atom);
      Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->level);
      Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->type);
      Tcl_DStringEndSublist(scheduleList);
      entryPtr = Tcl_NextHashEntry(&search);
    }
  }

  return(scheduleList);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleInfo --
 *
 *	This procedure returns a dynamically-allocated string  with
 *      overall information about a hash table, such as the number
 *      of entries it contains, the number of buckets in its  hash
 *      array,  and  the  utilization  of  the buckets.
 *
 * Results:
 *     A dynamically-allocated string, it is the  caller's responsibility 
 *     to free the result string by passing it to free.    
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

char *
BurtScheduleInfo(void)
{
    return(Tcl_HashStats(&scheduleTable));
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleLength --
 *
 *	This procedure returns the number of entries in the schedule table
 *
 * Results:
 *     An integer representing the size of the table
 *     
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int
BurtScheduleLength (void)
{
  return(BurtTableLength(&scheduleTable));
}

/*
void 
FormatUserOrderProcError(Tcl_Interp *interp, char *procName, char *message)
{
  Tcl_Obj *resultPtr;
  resultPtr = Tcl_NewStringObj("", -1);
  Tcl_AppendStringsToObj(resultPtr,
			 "BURT: ERROR in ordering proc, procedure:  ", 
			 procName, "\n",
			 "BURT: ERROR in ordering proc, message:    ",
			 message, "\n",
			 "BURT: ERROR in ordering proc, time:       ",
			 BurtTime(), "\n", NULL);
  Tcl_SetObjResult(interp, resultPtr);
}
*/


/*
 *----------------------------------------------------------------------
 *
 * GetScheduleVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetScheduleVersion(void)
{
  return(rcsid);
}
