/*
 * burtStatus.c
 *
 *    This file defines the status command.  It keeps track of the status of 
 *    the burt run.  Mostly it is the access point to the Burt_Status 
 *    structure.
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef lint
static char rcsid[] = "$Id: burtStatus.c,v 1.14 1999/04/01 20:54:31 jmelski Exp $";
#endif


#include <burtInt.h>

typedef struct Burt_Status {
  int runState;                  /* see burtInt.h for valid states */
  char tapeLabelStr[BURT_MAX_LABELLEN];      /* Tape label string. */ 
  unsigned int scheduled;        /* Number of atoms scheduled.     */
  unsigned int channels;         /* Number of channels running.    */
  unsigned int finished;         /* Number of atoms finished.      */
  unsigned int started;          /* Number of atoms started.       */
  unsigned int tapes;            /* Number of tapes used.          */
  unsigned int bytes;            /* Number of bytes written/read   */
  unsigned int xbytes;           /* Number of xbytes written/read  */
  time_t startTime;              /* Time operation began.          */
  time_t completeTime;           /* Time operation completed.      */
  
  unsigned int snapBytes;        /* Number of bytes since last snapshot */
  unsigned int snapXBytes;       /* Number of xbytes since last snapshot */
  time_t snapTime;               /* Time of last snapshot */
} Burt_Status;


static Burt_Status burtStatus;     /* Current status of burt */

/*
 *----------------------------------------------------------------------
 *
 * Burt_StatusCmd --
 *
 *	This procedure implements the Tcl command status.
 *
 * Results:
 *      Returns current status, or an error message if an error occurs
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

int 
Burt_StatusCmd(ClientData cdata, Tcl_Interp *interp, int argc, char **argv)
{
  int length;
  Tcl_DString *runningList;

  if (argc > 2)
  {
    Tcl_SetResult(interp, "wrong # args: should be \"status option\"", 
		  TCL_STATIC);
    return (TCL_ERROR);
  }

  if (argc == 1)
  {
    GetRuntimeStatus(interp);
    return(TCL_OK);
  }
  else 
  {
    length = strlen(argv[1]);
    if (strncmp(argv[1], "internalversion", length) == 0)
    {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,
		       GetBackupVersion(), "\n",
		       GetChecksumVersion(), "\n",
		       GetMainVersion(), "\n",
		       GetMtIOVersion(), "\n",
		       GetReadtapeVersion(), "\n",
		       GetRecoverVersion(), "\n",
		       GetScheduleVersion(), "\n",
		       GetStatusVersion(), "\n",
		       GetUtilVersion(), 
		       NULL);
      return(TCL_OK);
    }
    else if (strncmp(argv[1], "running", length) == 0)
    {
      switch (BurtStatusGetRunState() & OPMASK)
      {
	case BACKUP:
	  runningList = BurtBackupRunningList();
	  break;
	case RECOVER:
	  runningList = BurtRecoverRunningList();
	  break;
	case READTAPE:
	  runningList = BurtReadtapeRunningList();
	  break;
	default:
	  Tcl_SetResult(interp, "option not supported for current operation",
			TCL_STATIC);
	  return(TCL_ERROR);
	  break;
      }
      if (runningList == NULL)
      {
	Tcl_SetResult(interp, 
		      "Could not allocate memory for list!", 
		      TCL_STATIC);
	return(TCL_ERROR);
      } 
      else
      {
	Tcl_DStringResult(interp, runningList);
	Tcl_DStringFree(runningList);
	return(TCL_OK);
      }
    }
    else
    {
      Tcl_SetResult(interp, 
		    "invalid option: should be running or internalversion", 
		    TCL_STATIC);
      return(TCL_ERROR);
    }
  }
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusReset --
 *
 *	This procedure resets status in the burtStatus structure
 *      to initial start up values.
 *
 * Results:
 *
 * Side effects:
 *      burtStatus is cleared.  
 *----------------------------------------------------------------------
 */

void
BurtStatusReset(void)
{
  BurtStatusSetRunState(NONE | RUN);
  BurtStatusSetTapeLabel("None");
  BurtStatusSetScheduled(0);
  BurtStatusSetChannels(0);
  BurtStatusSetFinished(0);
  BurtStatusSetStarted(0);
  BurtStatusSetTapes(0);
  BurtStatusSetBytes(0);
  BurtStatusSetXBytes(0);
  BurtStatusSetStartTime(0);
  BurtStatusSetCompleteTime(0);
  BurtStatusSetSnapBytes(0);
  BurtStatusSetSnapXBytes(0);
  BurtStatusSetSnapTime(0);
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSave --
 *
 *      Creates a new Burt_Status structure and copies the current 
 *      burtStatus information into it.
 *
 * Results:
 *
 *      Returns a void pointer to the newly created structure
 *
 * Side effects:
 *
 *      Malloc's some memory which must be released eventually
 *----------------------------------------------------------------------
 */

char *
BurtStatusSave(void)
{
  Burt_Status *savedStatus = (Burt_Status *)(malloc(sizeof(Burt_Status)));
  if (savedStatus != NULL) 
    memcpy((void *)savedStatus, (void *)&burtStatus, sizeof(Burt_Status));
  return((char *)savedStatus);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusRestore --
 *
 *	Copies the data pointed to by *restoreStatus into the current 
 *      burtStatus structure.  
 *
 * Results:
 *     
 *      None
 *
 * Side effects:
 *   
 *      Overwrites current contents of the burtStatus structure
 *----------------------------------------------------------------------
 */

void
BurtStatusRestore(char *restoreStatus)
{
  if (restoreStatus != NULL) 
    memcpy((void *)&burtStatus, (void *)restoreStatus, sizeof(Burt_Status));
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusFree --
 *
 *	Frees the space pointed to by *freeStatus, assumed to be a Burt_Status
 *      structure previously allocated by a call to BurtStatusSave
 *
 * Results:
 *     
 *      None
 *
 * Side effects:
 *   
 *      Frees the memory pointed to by *freeStatus
 *----------------------------------------------------------------------
 */

void
BurtStatusFree(char *freeStatus)
{
  free((void *)freeStatus);
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetRunState --
 *
 *	This procedure set the burtStatus.runState value in
 *      the burtStatus structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetRunState(int runState)
{
  burtStatus.runState = runState;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetRunState --
 *
 *	This procedure returns runState of burtStatus structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int 
BurtStatusGetRunState(void)
{
  return(burtStatus.runState);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetTapeLabel --
 *
 *	This procedure sets the tapeLabel field of the burtStatus 
 *      structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetTapeLabel(char *label)
{
  strncpy(burtStatus.tapeLabelStr, label, BURT_MAX_LABELLEN);
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetTapeLabel --
 *
 *	This procedure returns the tapeLabel field of the burtStatus 
 *      structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

char *
BurtStatusGetTapeLabel(void)
{
  return(burtStatus.tapeLabelStr);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetScheduled --
 *
 *	This procedure sets the schduled field of the burtStatus.
 *      structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetScheduled(unsigned int numScheduled)
{
  burtStatus.scheduled = numScheduled;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetScheduled --
 *
 *	This procedure returns schduled field of the burtStatus.
 *      structure.  I.E The total number of schduled items in this run.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetScheduled(void)
{
  return(burtStatus.scheduled);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetChannels --
 *
 *	This procedure sets the channels field of the burtStatus
 *      structure.  I.E The number of channels in this run.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetChannels(unsigned int numChannels)
{
  burtStatus.channels = numChannels;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetChannels --
 *
 *	This procedure returns the channels field of the burtStatus
 *      structure.  I.E The number of channels in this run.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetChannels(void)
{
  return(burtStatus.channels);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetFinished --
 *
 *	This procedure sets the finished field of the burtStatus
 *      structure.  I.E The number of items finished in this run.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetFinished(unsigned int numFinished)
{
  burtStatus.finished = numFinished;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetFinished --
 *
 *	This procedure returns the finished field of the burtStatus
 *      structure.  I.E The number of items finished in this run. (NOT
 *      if Burt itself is finished!)
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetFinished(void)
{
  return(burtStatus.finished);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetStarted --
 *
 *	This procedure sets the started field of the burtStatus
 *      structure.  I.E The number of items started in this run. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetStarted(unsigned int numStarted)
{
  burtStatus.started = numStarted;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetStarted --
 *
 *	This procedure returns the started field of the burtStatus
 *      structure.  I.E The number of items started in this run. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetStarted(void)
{
  return(burtStatus.started);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetTapes --
 *
 *	This procedure sets the tapes field of the burtStatus
 *      structure.  I.E The number of tapes used in this run. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetTapes(unsigned int numTapes)
{
  burtStatus.tapes = numTapes;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetTapes --
 *
 *	This procedure returns number of tapes used during this backup.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetTapes(void)
{
  return(burtStatus.tapes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetBytes --
 *
 *	This procedure sets the bytes field of the burtStatus
 *      structure.  I.E The number of bytes written to tape in this run. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetBytes(unsigned int numBytes)
{
  burtStatus.bytes = numBytes;
  if (burtStatus.bytes >= xBYTES)
  {
    BurtStatusSetXBytes(BurtStatusGetXBytes() + (burtStatus.bytes / xBYTES));
    burtStatus.bytes = burtStatus.bytes % xBYTES;
  }
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetBytes --
 *
 *	This procedure returns number of bytes written during this backup.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetBytes(void)
{
  return(burtStatus.bytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetTotalXBytes --
 *
 *	This procedure returns total number of xbytes written, as a float
 *      This number is a combination of xbytes and bytes/xBYTES, so it is
 *      the floating point representation of the total written.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

float
BurtStatusGetTotalXBytes(void)
{
  float totalBytes;
  totalBytes = (float)BurtStatusGetXBytes();
  totalBytes += (float)BurtStatusGetBytes() / xBYTES;
  return(totalBytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetXBytes --
 *
 *	This procedure sets number of xbytes written (number of kilo- or
 *      mega- bytes depending on value of xBYTES)
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void
BurtStatusSetXBytes(unsigned int numXBytes)
{
  burtStatus.xbytes = numXBytes;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetXBytes --
 *
 *	This procedure returns number of xbytes written (number of kilo- or
 *      mega- bytes depeding on value of xBYTES)
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int
BurtStatusGetXBytes(void)
{
  return(burtStatus.xbytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetStartTime --
 *
 *	This procedure sets the startTime field of the burtStatus
 *      structure.  I.E The time this run started. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetStartTime(time_t startTime)
{
  burtStatus.startTime = startTime;
  burtStatus.snapTime = startTime;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetStartTime --
 *
 *	This procedure returns the startTime field of the burtStatus
 *      structure.  I.E The time this run started. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

time_t 
BurtStatusGetStartTime(void)
{
  return(burtStatus.startTime);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetCompleteTime --
 *
 *	This procedure sets the completeTime field of the burtStatus
 *      structure.  I.E The time this run completed. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetCompleteTime(time_t completeTime)
{
  burtStatus.completeTime = completeTime;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetCompleteTime --
 *
 *	This procedure returns the completeTime field of the burtStatus
 *      structure.  I.E The time this run completed. 
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

time_t 
BurtStatusGetCompleteTime(void)
{
  return(burtStatus.completeTime);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetSnapBytes --
 *
 *	This procedure sets the snapBytes field of the burtStatus
 *      structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetSnapBytes(unsigned int numBytes)
{
  burtStatus.snapBytes = numBytes;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetSnapXBytes --
 *
 *	This procedure sets number of snap sbytes. (number of kilo- or
 *      mega- bytes depending on value of xBYTES)
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetSnapXBytes(unsigned int numXBytes)
{
  burtStatus.snapXBytes = numXBytes;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetSnapBytes --
 *
 *	This procedure returns number of snap bytes.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetSnapBytes(void)
{
  return(burtStatus.snapBytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetSnapXBytes --
 *
 *	This procedure returns number of snap xbytes. (number of kilo- or
 *      mega- bytes depending on value of xBYTES)
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

unsigned int 
BurtStatusGetSnapXBytes(void)
{
  return(burtStatus.snapXBytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusSetSnapTime --
 *
 *	This procedure sets the snapTime field of the burtStatus
 *      structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

void 
BurtStatusSetSnapTime(time_t snapTime)
{
  burtStatus.snapTime = snapTime;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetSnapTime --
 *
 *	This procedure returns the snapTime field of the burtStatus
 *      structure.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

time_t 
BurtStatusGetSnapTime(void)
{
  return(burtStatus.snapTime);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetCurrentRate --
 *
 *	This procedure returns the current speed of the operation, ie
 *      the number of bytes written since the last check divided by the
 *      number of seconds since the last check.
 *
 * Results:
 *
 * Side effects:
 *      
 *      Updates the snapTime and snapBytes fields of the burtStatus structure
 *
 *----------------------------------------------------------------------
 */

float
BurtStatusGetCurrentRate(time_t now)
{
  float currentRate;
  float totalSnapXBytes;
  time_t timeDifference;

  if (now <= BurtStatusGetSnapTime())
    currentRate = 0.0;
  else
  {
    timeDifference = now - BurtStatusGetSnapTime();
    totalSnapXBytes = (float)BurtStatusGetSnapXBytes();
    totalSnapXBytes += (float)BurtStatusGetSnapBytes() / xBYTES;
    currentRate = BurtStatusGetTotalXBytes() - totalSnapXBytes;
    currentRate /= timeDifference;
    currentRate *= xBYTES;
  }
  BurtStatusSetSnapTime(now);
  BurtStatusSetSnapBytes(BurtStatusGetBytes());
  BurtStatusSetSnapXBytes(BurtStatusGetXBytes());
  return currentRate;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtStatusGetRate --
 *
 *	This procedure returns the overall speed of the operation, ie
 *      the number of bytes written since the start divided by the
 *      number of seconds since the start.
 *
 * Results:
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

float
BurtStatusGetRate(time_t now)
{
  float currentRate;
  time_t timeDifference;

  if (now <= BurtStatusGetStartTime())
    currentRate = 0.0;
  else
  {
    timeDifference = now - BurtStatusGetStartTime();
    currentRate = BurtStatusGetTotalXBytes() / timeDifference;
    currentRate *= xBYTES;
  }
  return currentRate;
}

/*
 *----------------------------------------------------------------------
 *
 * GetRuntimeStatus --
 *
 *	This procedure extracts status from the burtStatus structure
 *      and fills in the Tcl interp->result filed.
 *
 * Results:
 *
 * Side effects:
 *	interp's result field is filled in with the current runtime
 *      status of Burt.
 *----------------------------------------------------------------------
 */

void 
GetRuntimeStatus(Tcl_Interp *interp)
{
  time_t now;
  char statBuffer[80];

  now = time((time_t *)NULL);

  switch (BurtStatusGetRunState() & OPMASK) 
  {
    case NONE:
      sprintf(statBuffer, "runstate none");
      break;
    case BACKUP:
      sprintf(statBuffer, "runstate backup");
      break;
    case READTAPE: 
      sprintf(statBuffer, "runstate readtape");
      break;
    case RECOVER: 
      sprintf(statBuffer, "runstate recover");
      break;
    default: 
      sprintf(statBuffer, "runstate unknown");
      break;
  }

  Tcl_AppendResult(interp, statBuffer, NULL);
  
  switch (BurtStatusGetRunState() & STATEMASK)
  {
    case STOP:
      sprintf(statBuffer, "-stop ");
      break;
    case COMPLETE:
      sprintf(statBuffer, "-complete ");
      break;
    case ABORT:
      sprintf(statBuffer, "-abort ");
      break;
    case WAITING:
      sprintf(statBuffer, "-waiting ");
      break;
    case RUN:
    default:
      sprintf(statBuffer, " ");
      break;
  }
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "tapelabel %s ", BurtStatusGetTapeLabel());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "scheduled %u ", BurtStatusGetScheduled());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "started %u ", BurtStatusGetStarted());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "channels %u ", BurtStatusGetChannels());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "finished %u ", BurtStatusGetFinished());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "tapes %u ", BurtStatusGetTapes());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "megabytes %1.2f ", BurtStatusGetTotalXBytes());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "starttime %d ", (int) BurtStatusGetStartTime());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "completetime %d ", (int) BurtStatusGetCompleteTime());
  Tcl_AppendResult(interp, statBuffer, NULL);
  
  sprintf(statBuffer, "currenttime %d ", (int) now);
  Tcl_AppendResult(interp, statBuffer, NULL);

  sprintf(statBuffer, "rate %f ", BurtStatusGetRate(now));
  Tcl_AppendResult(interp, statBuffer, NULL);

  sprintf(statBuffer, "currentrate %f ", BurtStatusGetCurrentRate(now));
  Tcl_AppendResult(interp, statBuffer, NULL);
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * GetStatusVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetStatusVersion(void)
{
  return(rcsid);
}
