/*
 * burtUtil.c
 *
 *    This file implements utilities used by more than one module in Burt
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the 
 * software.  A reasonable copying fee may be charged for any distribution of 
 * this software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no 
 * representations about the suitability of this software for any purpose.  
 * It is provided "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifndef lint
static char rcsid[] = "$Id: burtUtil.c,v 1.11 1999/04/15 18:42:23 jmelski Exp $";
#endif

#include <burtInt.h>
static void ProcessLoadWritableTapeTimer(ClientData clientData);
static void ProcessLoadReadableTapeTimer(ClientData clientData);

/*
 *----------------------------------------------------------------------
 *
 * BurtTableLength --
 *
 *	This procedure returns the number of items in the table.  There 
 *      two ways to do this:  first, you could walk the table and count the
 *      entries for yourself; second, you could peek into the Tcl_HashTable
 *      structure and grab the size from that.  The second is much faster,
 *      of course, but it's probably not "good" behaviour (though the core
 *      designers should have included this functionality themselves).  Until
 *      I hear that I'm not supposed to peek into the hash table like this,
 *      I'll stick with the faster method, but I will leave the slower method
 *      in place, should it ever be needed again.
 *
 * Results:
 *	Returns the number of items in the table.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int
BurtTableLength(Tcl_HashTable *tablePtr)
{
/*
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch search;
    int length;

    length = 0;
    entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
    while (entryPtr != NULL) {
	length++;
	entryPtr = Tcl_NextHashEntry(&search);
    }

    return(length);
*/
  return(tablePtr->numEntries);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtTime --
 *
 *      Generates a standard Burt timestamp in some static memory. 
 *      
 * Results:
 *      
 *      Returns a pointer to the timestamp.
 *	
 * Side effects:
 *      
 *      Wipes out whatever existed at the spot before.
 *      
 *----------------------------------------------------------------------
 */

char *
BurtTime(void)
{
  time_t timer;
  static char now[32];

  timer = time((time_t *)NULL);
  strftime(now, 31, "{%X %x}", localtime(&timer));
  return(now);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtSetupHashKey --
 *
 *      Given a host, atom, level and type, sets up a hash key
 *      
 * Side effects:
 *      
 *      Overwrites the memory pointed to by *key
 *      
 *----------------------------------------------------------------------
 */

void 
BurtSetupHashKey (char *key, 
		  char *host, char *atom, char *level, char *type)
{
  strncpy(key, host, BURT_MAX_HOSTLEN);
  strcat(key, atom);
  strcat(key, level);
  strcat(key, type);
}


/*
 *----------------------------------------------------------------------
 *
 * LoadNextTape --
 *
 *      Loads the next tape, ensuring that it is readable or writable
 *      depending on the value of mode; uses prompt and logChannel to 
 *      display prompts to the user.
 *      
 * Results:
 *      TCL_OK if a tape was loaded; TCL_ERROR if there were any problems
 *
 * Side effects:
 *      Loads a new tape
 *      
 *----------------------------------------------------------------------
 */
int 
LoadNextTape(Tcl_Channel mtChannel, Tcl_Channel logChannel, 
	     char *prompt, int mode)
{
  int foundEvent;
  Burt_LoadNextTapeData promptData;

  /* Try to rewind and eject the tape */
  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return (TCL_ERROR);
  
  if (BurtMtSafeOffline(mtChannel) == MT_ERROR) 
    return (TCL_ERROR);

  Tcl_Write(logChannel, prompt, -1);
  Tcl_Write(logChannel, " ", -1);
  Tcl_Write(logChannel, BurtTime(), -1);
  Tcl_Write(logChannel, "\n", -1);

  /* set up the Burt_LoadNextTapeData structure */
  promptData.mtChannel  = mtChannel;
  promptData.logChannel = logChannel;
  promptData.tapeReady  = 0;
  promptData.prompt     = prompt;
  if ((mode & TCL_WRITABLE) == TCL_WRITABLE)
  {
    promptData.timer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000,
					      ProcessLoadWritableTapeTimer,
					      (ClientData) &promptData);
  } 
  else if ((mode & TCL_READABLE) == TCL_READABLE)
  {
    promptData.timer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000,
					      ProcessLoadReadableTapeTimer,
					      (ClientData) &promptData);
  }
  else
  {
    Tcl_Write(logChannel, "BURT: ERROR in LoadNextTape, invalid mode ", -1);
    Tcl_Write(logChannel, BurtTime(), -1);
    Tcl_Write(logChannel, "\n", -1);
    return(TCL_ERROR);
  }

  /* Keep the event loop going */
  foundEvent = 1;
  while (!promptData.tapeReady && foundEvent)
    foundEvent = Tcl_DoOneEvent(0);

  Tcl_DeleteTimerHandler(promptData.timer);

  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessLoadWritableTapeTimer --
 *
 *      Checks if a writable tape has been loaded; if not, registers a 
 *      timer event to check again after TAPE_LOAD_DELAY seconds.
 *      
 * Side effects:
 *      May register a timer event; on success, writes a label to the tape
 *      
 *----------------------------------------------------------------------
 */
static void
ProcessLoadWritableTapeTimer(ClientData clientData)
{
  Burt_LoadNextTapeData *data = (Burt_LoadNextTapeData *)clientData;
  Burt_TapeLabel tapeLabel;
  int nbytes;

  /* If we can rewind the tape and write a label to it, it is considered OK */
  if (BurtMtRewind(data->mtChannel) == MT_OK)
  {
    SetupTapeLabel(&tapeLabel, BurtStatusGetTapeLabel(), BurtStatusGetTapes());
    nbytes = BurtMtWriteLabel(data->mtChannel, &tapeLabel);
    if (nbytes != MT_ERROR)
    {
      data->tapeReady = 1;
      BurtStatusSetBytes(BurtStatusGetBytes() + nbytes);
    }
  }
  
  /* If the tape is not yet ready, print the prompt and reset the timer */
  if (!data->tapeReady)
  {
    Tcl_Write(data->logChannel, data->prompt, -1);
    Tcl_Write(data->logChannel, " ", -1);
    Tcl_Write(data->logChannel, BurtTime(), -1);
    Tcl_Write(data->logChannel, "\n", -1);
    data->timer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
					 ProcessLoadWritableTapeTimer,
					 clientData);
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessLoadReadableTapeTimer --
 *
 *      Checks if a readable tape has been inserted, and if it is the correct
 *      tape; if not, registers a timer event to check again after 
 *      TAPE_LOAD_DELAY seconds.
 *      
 * Side effects:
 *      May register a timer event
 *      
 *----------------------------------------------------------------------
 */
static void
ProcessLoadReadableTapeTimer(ClientData clientData)
{
  Burt_LoadNextTapeData *data = (Burt_LoadNextTapeData *)clientData;
  Burt_TapeLabel actualLabel, expectedLabel;
  char buffer[1024];

  /* 
   * If we can rewind the tape, read the label, and verify that the label is
   * what we expect it to be, the tape is considered OK 
   */
  if (BurtMtRewind(data->mtChannel) == MT_OK)
  {
    if (BurtMtReadLabel(data->mtChannel, &actualLabel) == MT_OK)
    {
      /* Make sure the correct tape has been inserted */
      SetupTapeLabel(&expectedLabel, 
		     BurtStatusGetTapeLabel(), 
		     BurtStatusGetTapes());
      if (TestTapeLabelEquality(&expectedLabel, &actualLabel) == 1)
      {
	data->tapeReady = 1;
      }
      else
      {
	sprintf(buffer, "BURT: ERROR, expected tape %s, #%s; got %s, #%s %s\n",
		expectedLabel.label, 
		expectedLabel.tape, 
		actualLabel.label, 
		actualLabel.tape, 
		BurtTime());
	Tcl_Write(data->logChannel, buffer, -1);
	BurtMtOffline(data->mtChannel);
      }
    }
  }

  /* If the tape is not yet ready, print the prompt and reset the timer */
  if (!data->tapeReady)
  {
    Tcl_Write(data->logChannel, data->prompt, -1);
    Tcl_Write(data->logChannel, " ", -1);
    Tcl_Write(data->logChannel, BurtTime(), -1);
    Tcl_Write(data->logChannel, "\n", -1);
    data->timer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
					 ProcessLoadReadableTapeTimer,
					 clientData);
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * SetupTapeLabel --
 *
 *	Just like the name says, this function sets up a Burt_TapeLabel
 *      structure with the current information.
 *
 * Side effects:
 *      The Burt_TapeLabel struct pointed to by tapeLabel will get its 
 *      members overwritten with the current information.
 *----------------------------------------------------------------------
 */

void
SetupTapeLabel(Burt_TapeLabel *tapeLabel, char *label, unsigned int tape)
{
  time_t now;
  char *dateStr;
  
  tapeLabel->magic = BURT_MAGIC_NUMBER;
  strncpy(tapeLabel->label, label, BURT_MAX_LABELLEN);
  now = time((time_t *)NULL);
  dateStr = ctime(&now);
  dateStr[strlen(dateStr) - 1] = '\0';
  strncpy(tapeLabel->date, dateStr, BURT_MAX_DATELEN);
  sprintf(tapeLabel->tape, "%u", tape);
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * TestTapeLabelEquality --
 *
 *	Verifies that the the two given tape labels are the same, with 
 *      respect to the label, tape number and magic number.
 *
 * Results:
 *      1 if the labels are the same, 0 otherwise.
 *----------------------------------------------------------------------
 */
int
TestTapeLabelEquality(Burt_TapeLabel *labelA, Burt_TapeLabel *labelB)
{
  /* Check tape label */
  if (strncmp(labelA->label, labelB->label, BURT_MAX_LABELLEN) != 0)
    return(0);

  /* Check tape number */
  if (strncmp(labelA->tape, labelB->tape, 3) != 0)
    return(0);

  /* Don't check magic number; due to the way magic numbers are stored 
   * internally, the magic number test will always fail when reading tapes
   * written by old versions of Burt
   */

  return(1);
}


/*
 *----------------------------------------------------------------------
 *
 * GetUtilVersion
 *     Returns a pointer to the rcsid string
 *
 * Results:
 *     A pointer to the rcsid string
 *
 *----------------------------------------------------------------------
 */
char *
GetUtilVersion(void)
{
  return(rcsid);
}

int GetMaxPacketSize()
{
  return(BURT_MAX_BLOCKSIZE - 580);
}
